# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from ai_flow.util.json_utils import Jsonable
from typing import Text, Dict, List
from datetime import datetime, timedelta
from pytz import timezone


class PeriodicConfig(Jsonable):
    """
    Define the periodic running configuration of the running unit
    (ai_flow.workflow.job.Job and ai_flow.workflow.workflow.Workflow).
    """

    def __init__(self,
                 trigger_config: Dict) -> None:
        """
        :param trigger_config: Support two types of configuration:
        1. cron config: {'start_date': 'start_date_expression', 'cron': 'cron_expression'}
        start_date_expression:
        year:int,month:int,day:int,hour:int,minute:int,second:int,Option[tzinfo: str]
        cron_expression:
        seconds minutes hours days months weeks years
        2. interval config {'start_date': 'start_date_expression', 'interval': 'interval_expression'}
        start_date_expression:
        year:int,month:int,day:int,hour:int,minute:int,second:int,Option[tzinfo: str]
        interval_expression:
        days:int,hours:int,minutes:int,seconds:int
        """
        super().__init__()
        self.trigger_config: Dict = trigger_config

    @classmethod
    def to_dict(cls, config: 'PeriodicConfig') -> Dict:
        if 'cron' in config.trigger_config:
            return {'start_date': config.trigger_config.get('start_date'),
                    'cron': config.trigger_config.get('cron')}
        elif 'interval' in config.trigger_config:
            return {'start_date': config.trigger_config.get('start_date'),
                    'interval': config.trigger_config.get('interval')}
        else:
            raise Exception('Periodic config must be one of:\n'
                            """1. cron config: {'start_date': 'start_date_expression', 'cron': 'cron_expression'}\n"""
                            """2. interval config {'start_date': 'start_date_expression', 'interval': 'interval_expression'}""")

    @classmethod
    def from_dict(cls, data: Dict) -> 'PeriodicConfig':
        return PeriodicConfig(trigger_config=data)

    def get_cron_items(self):
        cron_list = self.trigger_config.get('cron').split(' ')
        if len(cron_list) != 7:
            raise Exception('cron expression {} is not validated! '
                            'Usage: seconds minutes hours days months weeks years')
        result = []
        for i in cron_list:
            result.append(i.strip())
        return result

    def get_start_date_items(self):
        start_date_list = self.trigger_config.get('start_date').split(',')
        if len(start_date_list) == 7:
            result = []
            for i in range(len(start_date_list)):
                if i < 6:
                    if len(start_date_list[i].strip()) == 0:
                        if i < 3:
                            raise Exception('year month, day mast set!')
                        else:
                            result.append(0)
                    else:
                        result.append(int(start_date_list[i].strip()))
                else:
                    if len(start_date_list[i].strip()) == 0:
                        result.append(None)
                    else:
                        result.append(start_date_list[i].strip())
            return result
        elif len(start_date_list) == 6:
            result = []
            for i in start_date_list:
                result.append(int(i.strip()))
            return result
        else:
            raise Exception('start expression {} is not validated! '
                            'Usage: year:int,month:int,day:int,hour:int,minute:int,second:int,Option[tzinfo: str]')

    def get_interval_items(self) -> List[float]:
        interval_list = self.trigger_config.get('interval').split(',')
        if len(interval_list) != 4:
            raise Exception('interval expression {} is not validated! '
                            'Usage: days:float,hours:float,minutes:float,seconds:float')
        result = []
        for i in interval_list:
            if len(i.strip()) == 0:
                result.append(0)
            else:
                result.append(int(i.strip()))
        return result

    def get_start_date(self) -> datetime:
        tmp = self.get_start_date_items()
        if tmp[6] is None:
            return datetime(year=tmp[0],
                            month=tmp[1],
                            day=tmp[2],
                            hour=tmp[3],
                            minute=tmp[4],
                            second=tmp[5])
        else:
            return datetime(year=tmp[0],
                            month=tmp[1],
                            day=tmp[2],
                            hour=tmp[3],
                            minute=tmp[4],
                            second=tmp[5],
                            tzinfo=timezone(tmp[6]))

    def get_interval(self) -> timedelta:
        tmp = self.get_interval_items()
        return timedelta(days=tmp[0],
                         hours=tmp[1],
                         minutes=tmp[2],
                         seconds=tmp[3])



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Main executable module"""

import os

import argcomplete

from airflow.cli import cli_parser
from airflow.configuration import conf


def main():
    """Main executable function"""
    if conf.get("core", "security") == 'kerberos':
        os.environ['KRB5CCNAME'] = conf.get('kerberos', 'ccache')
        os.environ['KRB5_KTNAME'] = conf.get('kerberos', 'keytab')

    parser = cli_parser.get_parser()
    argcomplete.autocomplete(parser)
    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This module contains an operator to run downstream tasks only for the
latest scheduled DagRun
"""
from typing import Dict, Iterable, Union

import pendulum

from airflow.operators.branch import BaseBranchOperator


class LatestOnlyOperator(BaseBranchOperator):
    """
    Allows a workflow to skip tasks that are not running during the most
    recent schedule interval.

    If the task is run outside of the latest schedule interval (i.e. external_trigger),
    all directly downstream tasks will be skipped.

    Note that downstream tasks are never skipped if the given DAG_Run is
    marked as externally triggered.
    """

    ui_color = '#e9ffdb'  # nyanza

    def choose_branch(self, context: Dict) -> Union[str, Iterable[str]]:
        # If the DAG Run is externally triggered, then return without
        # skipping downstream tasks
        if context['dag_run'] and context['dag_run'].external_trigger:
            self.log.info("Externally triggered DAG_Run: allowing execution to proceed.")
            return list(context['task'].get_direct_relative_ids(upstream=False))

        now = pendulum.now('UTC')
        left_window = context['dag'].following_schedule(context['execution_date'])
        right_window = context['dag'].following_schedule(left_window)
        self.log.info(
            'Checking latest only with left_window: %s right_window: %s now: %s',
            left_window,
            right_window,
            now,
        )

        if not left_window < now <= right_window:
            self.log.info('Not latest execution, skipping downstream.')
            # we return an empty list, thus the parent BaseBranchOperator
            # won't exclude any downstream tasks from skipping.
            return []
        else:
            self.log.info('Latest, allowing execution to proceed.')
            return list(context['task'].get_direct_relative_ids(upstream=False))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from airflow.settings import STATE_COLORS


class State:
    """
    Static class with task instance states constants and color method to
    avoid hardcoding.
    """

    # scheduler
    NONE = None  # type: None
    REMOVED = "removed"
    SCHEDULED = "scheduled"

    # set by the executor (t.b.d.)
    # LAUNCHED = "launched"

    # set by a task
    QUEUED = "queued"
    RUNNING = "running"
    SUCCESS = "success"
    SHUTDOWN = "shutdown"  # External request to shut down
    FAILED = "failed"
    UP_FOR_RETRY = "up_for_retry"
    UP_FOR_RESCHEDULE = "up_for_reschedule"
    UPSTREAM_FAILED = "upstream_failed"
    SKIPPED = "skipped"
    SENSING = "sensing"
    KILLING = "killing"
    KILLED = "killed"

    task_states = (
        SUCCESS,
        RUNNING,
        FAILED,
        UPSTREAM_FAILED,
        SKIPPED,
        UP_FOR_RETRY,
        UP_FOR_RESCHEDULE,
        QUEUED,
        NONE,
        SCHEDULED,
        SENSING,
        KILLED,
        KILLING,
    )

    dag_states = (
        SUCCESS,
        RUNNING,
        FAILED,
    )

    state_color = {
        QUEUED: 'gray',
        RUNNING: 'lime',
        SUCCESS: 'green',
        SHUTDOWN: 'blue',
        FAILED: 'red',
        UP_FOR_RETRY: 'gold',
        UP_FOR_RESCHEDULE: 'turquoise',
        UPSTREAM_FAILED: 'orange',
        SKIPPED: 'pink',
        REMOVED: 'lightgrey',
        SCHEDULED: 'tan',
        NONE: 'lightblue',
        SENSING: 'lightseagreen',
        KILLED: 'red',
    }
    state_color.update(STATE_COLORS)  # type: ignore

    @classmethod
    def color(cls, state):
        """Returns color for a state."""
        return cls.state_color.get(state, 'white')

    @classmethod
    def color_fg(cls, state):
        """Black&white colors for a state."""
        color = cls.color(state)
        if color in ['green', 'red']:
            return 'white'
        return 'black'

    running = frozenset([RUNNING, SENSING])
    """
    A list of states indicating that a task is being executed.
    """

    finished = frozenset(
        [
            SUCCESS,
            FAILED,
            SKIPPED,
            UPSTREAM_FAILED,
            KILLED,
        ]
    )
    """
    A list of states indicating a task has reached a terminal state (i.e. it has "finished") and needs no
    further action.

    Note that the attempt could have resulted in failure or have been
    interrupted; or perhaps never run at all (skip, or upstream_failed) in any
    case, it is no longer running.
    """

    unfinished = frozenset(
        [
            NONE,
            SCHEDULED,
            QUEUED,
            RUNNING,
            SENSING,
            SHUTDOWN,
            UP_FOR_RETRY,
            UP_FOR_RESCHEDULE,
            KILLING,
        ]
    )
    """
    A list of states indicating that a task either has not completed
    a run or has not even started.
    """

    failed_states = frozenset([FAILED, UPSTREAM_FAILED, KILLED])
    """
    A list of states indicating that a task or dag is a failed state.
    """

    success_states = frozenset([SUCCESS, SKIPPED])
    """
    A list of states indicating that a task or dag is a success state.
    """


class PokeState:
    """Static class with poke states constants used in smart operator."""

    LANDED = 'landed'
    NOT_LANDED = 'not_landed'
    POKE_EXCEPTION = 'poke_exception'



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""add task fails journal table

Revision ID: 64de9cddf6c9
Revises: 211e584da130
Create Date: 2016-08-03 14:02:59.203021

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
from airflow.models.base import COLLATION_ARGS

revision = '64de9cddf6c9'
down_revision = '211e584da130'
branch_labels = None
depends_on = None


def upgrade():  # noqa: D103
    op.create_table(
        'task_fail',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('task_id', sa.String(length=250, **COLLATION_ARGS), nullable=False),
        sa.Column('dag_id', sa.String(length=250, **COLLATION_ARGS), nullable=False),
        sa.Column('execution_date', sa.DateTime(), nullable=False),
        sa.Column('start_date', sa.DateTime(), nullable=True),
        sa.Column('end_date', sa.DateTime(), nullable=True),
        sa.Column('duration', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
    )


def downgrade():  # noqa: D103
    op.drop_table('task_fail')



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import unittest

from ai_flow.util import json_utils
from ai_flow.graph.edge import Edge


class TestEdge(unittest.TestCase):

    def test_edge_serde(self):
        edge = Edge("a", 'b')
        json_text = json_utils.dumps(edge)
        c2: Edge = json_utils.loads(json_text)
        self.assertEqual(edge.source, c2.source)


if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import threading
from concurrent.futures import Future
from concurrent.futures.thread import ThreadPoolExecutor
from queue import Queue, Empty
from typing import Optional, Set, Dict, cast

from airflow.utils.session import create_session
from notification_service.base_notification import BaseEvent

from airflow.contrib.jobs.background_service import BackgroundService
from airflow.executors.scheduling_action import SchedulingAction
from airflow.models import BaseOperator, DagRun
from airflow.models.dag import DagEventDependencies
from airflow.models.eventhandler import EventKey
from airflow.models.serialized_dag import SerializedDagModel
from airflow.models.taskstate import TaskState
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.mailbox import Mailbox
from airflow.events.scheduler_events import EventHandleEvent


class DagRunId(object):
    """
    DagRunId include dag id and run id of a dag run that is necessary to uniquely identify a dag run of a dag.
    """

    def __init__(self, dag_id: str, run_id: str) -> None:
        self.dag_id = dag_id
        self.run_id = run_id

    def __eq__(self, o: object) -> bool:
        if type(o) is type(self):
            o = cast(DagRunId, o)
            return self.dag_id == o.dag_id and self.run_id == o.run_id
        return False

    def __hash__(self) -> int:
        return hash(self.dag_id) + hash(self.run_id)

    def __str__(self):
        return str(self.__dict__)


class EventHandleResult(object):
    """
    :class:`EventHandleResult` contains dag_run_id, task_id, and scheduling_action. It is sent back to the
    scheduler after executing the event_handler for the task.
    """

    def __init__(self, dag_run_id: DagRunId, task_id: str, scheduling_action: SchedulingAction):
        self.dag_run_id = dag_run_id
        self.task_id = task_id
        self.scheduling_action = scheduling_action

    def __eq__(self, o: object) -> bool:
        return self.__dict__ == o.__dict__

    def to_event(self) -> EventHandleEvent:
        return EventHandleEvent(dag_id=self.dag_run_id.dag_id,
                                dag_run_id=self.dag_run_id.run_id,
                                task_id=self.task_id,
                                action=self.scheduling_action)

    @classmethod
    def from_event(cls, event: EventHandleEvent) -> 'EventHandleResult':
        return EventHandleResult(DagRunId(dag_id=event.dag_id, run_id=event.dag_run_id),
                                 task_id=event.task_id,
                                 scheduling_action=event.action)


class DagRunEventManager(BackgroundService, LoggingMixin):
    """
    :class:`DagRunEventManager` accepts event and the dag_run_id of the dag run to handle the event from the
    Scheduler. It manage the life cycle of the :class:`DagRunEventExecutorRunner`.
    """

    def __init__(self, mailbox: Mailbox, context=None, max_workers=None, max_num_event=16):
        super().__init__(context)
        self._mailbox = mailbox
        self._event_executor_runners: Dict[DagRunId, DagRunEventExecutorRunner] = dict()
        self._executor = ThreadPoolExecutor(max_workers=max_workers, thread_name_prefix="DagRunEventExecutor")
        self._max_num_event = max_num_event

        # DagRunEventExecutor will be added to the set right after it is submitted to ExecutorPool and removed
        # right after it is finished running.
        self._running_event_executor_runners: Set[DagRunId] = set()
        self._lock = threading.RLock()

    def start(self):
        """
        Do nothing
        """
        pass

    def handle_event(self, dag_run_id: DagRunId, event: BaseEvent):
        """
        Handle the event for :class:`DagRun` with the given dag run id.
        :param dag_run_id: id of the dagrun.
        :type dag_run_id: DagRunId
        :param event: the event to be handled.
        :type event: BaseEvent
        """
        with self._lock:
            # create DagRunEventExecutorRunner if not exist
            if dag_run_id not in self._event_executor_runners:
                self._event_executor_runners[dag_run_id] = \
                    DagRunEventExecutorRunner(mailbox=self._mailbox,
                                              dag_run_id=dag_run_id,
                                              max_num_event=self._max_num_event)
            self._event_executor_runners[dag_run_id].put_event(event)

            # submit the DagRunEventExecutorRunner to ExecutorPool if it is not running
            if dag_run_id in self._running_event_executor_runners:
                self.log.debug("{} has running runner".format(dag_run_id))
                return

            self._submit_event_executor_runner(dag_run_id)

    def end(self) -> None:
        """
        Stops all the :class:`DagRunEventExecutorRunner` and wait util all of them terminated.
        """
        self._executor.shutdown(wait=True)

    def terminate(self):
        """
        Stops all the :class:`DagRunEventExecutorRunner` and return immediately.
        """
        self._executor.shutdown(wait=False)

    def is_alive(self) -> bool:
        return True

    def _submit_event_executor_runner(self, dag_run_id):
        self._running_event_executor_runners.add(dag_run_id)
        self._executor.submit(self._start_event_executor_runner, dag_run_id) \
            .add_done_callback(self._runner_exited)

    def _start_event_executor_runner(self, dag_run_id: DagRunId):
        self.log.debug("starting DagRunEventExecutorRunner for {}".format(dag_run_id))
        self._event_executor_runners[dag_run_id].run()
        return dag_run_id

    def _runner_exited(self, dag_run_id: Future):
        dag_run_id = dag_run_id.result()
        self.log.debug("DagRunEventExecutorRunner for {} exit".format(dag_run_id))
        with self._lock:
            if not self._event_executor_runners[dag_run_id].empty_event_queue():
                self.log.debug("DagRunEventExecutorRunner for {} has non empty event queue, resubmitting...")
                self._submit_event_executor_runner(dag_run_id)
                return
            self._running_event_executor_runners.remove(dag_run_id)
            # DagRunEventExecutorRunner finished with empty event queue, release the DagRunEventExecutorRunner
            self._event_executor_runners.pop(dag_run_id)


class DagRunEventExecutorRunner(LoggingMixin):
    """
    :class:`DagRunEventExecutorRunner` use the :class:`DagRunEventExecutor` to run the event handler for events.
    """

    def __init__(self, mailbox: Mailbox,
                 dag_run_id: DagRunId,
                 max_num_event: int,
                 poll_timeout: int = 0) -> None:
        """

        :param mailbox: where the EventHandleResult is send to.
        :type mailbox: Mailbox
        :param dag_run_id: the run id of the dag run
        :type dag_run_id: str
        :param max_num_event: max number of event can be handled before exit
        :type max_num_event: int
        :param poll_timeout: poll timeout in second for event before exit
        :type poll_timeout: int
        """
        super().__init__()
        self._mailbox = mailbox
        self._dag_run_id = dag_run_id
        self._event_queue = Queue()
        self._max_num_event = max_num_event
        self._poll_timeout = poll_timeout

        dag_runs = DagRun.find(dag_id=dag_run_id.dag_id, run_id=dag_run_id.run_id)
        if len(dag_runs) < 1:
            raise RuntimeError("no dag_run found with dag_run_id: {}".format(dag_run_id))
        elif len(dag_runs) > 1:
            raise RuntimeError("more than one dag_run found with dag_run_id: {}".format(dag_run_id))

        self._dag_run = dag_runs[0]
        dag_id = self._dag_run.dag_id
        self._serialized_dag_model = SerializedDagModel.get(dag_id)
        if self._serialized_dag_model is None:
            raise RuntimeError("no serialized dag is found with dag_id: {}".format(dag_id))

    def run(self) -> None:
        """
        Start handling events. It will return in the following conditions:
        1. It doesn't receive any event in some timeout specified by poll_timeout.
        2. It has handled some number of event specified by max_num_event
        """
        try:
            self._run()
            self.log.debug("DagRunEventExecutorRunner for {} exiting...".format(self._dag_run_id))
        except Exception as _:
            self.log.exception("Failed to run event executor for dag run {}: ".format(self._dag_run_id))

    def put_event(self, event: BaseEvent):
        """
        Put the event to be handled by :class:`DagRunEventExecutor`.
        :param event: event to be handled.
        :type event:
        """
        self.log.debug("putting event {} to DagRunEventExecutorRunner {}".format(event, self._dag_run_id))
        self._event_queue.put(event)

    def empty_event_queue(self) -> bool:
        """
        Check whether event queue is empty.
        :return:
        :rtype:
        """
        return self._event_queue.qsize() == 0

    def _run(self):
        executor = DagRunEventExecutor(self._serialized_dag_model)
        event_processed = 0
        while True:
            event = self._get_event()
            self.log.debug("DagRunEventExecutorRunner for {} got event {}".format(self._dag_run_id, event))
            if event is None:
                break
            for task_id, action in executor.execute_event_handler(self._dag_run, event).items():
                self._mailbox.send_message(EventHandleResult(self._dag_run_id, task_id, action).to_event().to_event())
            event_processed = event_processed + 1
            if event_processed >= self._max_num_event:
                self.log.debug("DagRunEventExecutorRunner for {} exceed max_num_event: {}"
                               .format(self._dag_run_id, self._max_num_event))
                break

    def _get_event(self, block: bool = True) -> Optional[BaseEvent]:
        self.log.debug("getting event")
        try:
            return self._event_queue.get(block=block, timeout=self._poll_timeout)
        except Empty as _:
            self.log.debug("no event found")
            return None


class DagRunEventExecutor(LoggingMixin):
    """
    :class:`DagRunEventExecutor` is used to execute the EventHandler for Tasks in one dag.
    """

    def __init__(self, serialized_dag_model: SerializedDagModel):
        """
        :param serialized_dag_model: the dag where the tasks that should run event handler belongs to.
        :type serialized_dag_model: SerializedDagModel
        """
        super().__init__()
        self._dag = serialized_dag_model.dag
        self._dependency: DagEventDependencies = \
            DagEventDependencies.from_json(serialized_dag_model.event_relationships)

    def execute_event_handler(self,
                              dag_run: DagRun,
                              event: BaseEvent) -> Dict[str, SchedulingAction]:
        """
        Execute the task event handler that are interested in the given event for a specific dag run.
        :param dag_run: dag run for which the event handler should execute.
        :type dag_run: DagRun
        :param event: the event to be handled.
        :type event: BaseEvent
        """
        task_ids = self._dependency.find_affected_tasks(EventKey(event.key,
                                                                 event.event_type,
                                                                 event.namespace,
                                                                 event.sender))
        if task_ids is None:
            return {}
        operators: Set[BaseOperator] = set()
        for task_id in task_ids:
            operator = self._find_operator(task_id)
            if operator is None:
                raise RuntimeError("Operator with task_id: {} not found in dag: {}"
                                   .format(task_id, self._dag.dag_id))
            operators.add(operator)

        scheduling_actions = dict()
        for operator in operators:
            scheduling_actions[operator.task_id] = self._operator_handle_event(event, operator, dag_run.execution_date)
        self.log.info("DagRunEventExecutor for dag {}-{} handle event: {} actions: {}"
                      .format(dag_run.dag_id, dag_run.run_id, event, scheduling_actions))
        return scheduling_actions

    @staticmethod
    def _operator_handle_event(event, operator, execution_date) -> SchedulingAction:
        task_state = TaskState.get_task_state(operator.dag_id, operator.task_id, execution_date)
        event_handler = operator.get_events_handler()
        if not task_state:
            task_state = TaskState(task_id=operator.task_id, dag_id=operator.dag_id, execution_date=execution_date)
            with create_session() as session:
                session.add(task_state)
                session.commit()
        scheduling_action, state = event_handler.handle_event(event, task_state.task_state)
        task_state.task_state = state
        task_state.update_task_state()
        return scheduling_action

    def _find_operator(self, task_id) -> Optional[BaseOperator]:
        if self._dag.has_task(task_id):
            return self._dag.get_task(task_id)
        return None



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from ai_flow.meta.artifact_meta import ArtifactMeta
from ai_flow.meta.dataset_meta import DatasetMeta, List, DataType, Schema
from ai_flow.meta.job_meta import State
from ai_flow.meta.model_meta import ModelMeta, ModelVersionMeta
from ai_flow.meta.model_relation_meta import ModelRelationMeta, ModelVersionRelationMeta
from ai_flow.meta.project_meta import ProjectMeta
from ai_flow.meta.workflow_meta import WorkflowMeta
from ai_flow.protobuf.message_pb2 import DatasetProto, ProjectProto, \
    StateProto, ModelRelationProto, ModelVersionRelationProto, ModelProto, ModelVersionProto, \
    ArtifactProto, ModelVersionStage, DataTypeProto, WorkflowMetaProto


class ProtoToMeta:

    @staticmethod
    def proto_to_dataset_meta(dataset_proto: DatasetProto) -> DatasetMeta:
        properties = dataset_proto.properties
        if properties == {}:
            properties = None
        name_list = dataset_proto.schema.name_list
        if not name_list:
            name_list = None
        type_list = dataset_proto.schema.type_list
        if not type_list:
            data_type_list = None
        else:
            data_type_list = []
            for c in type_list:
                data_type_list.append(DataType(DataTypeProto.Name(c)))
        data_format = dataset_proto.data_format.value if dataset_proto.HasField('data_format') else None
        description = dataset_proto.description.value if dataset_proto.HasField('description') else None
        uri = dataset_proto.uri.value if dataset_proto.HasField('uri') else None
        create_time = dataset_proto.create_time.value if dataset_proto.HasField('create_time') else None
        update_time = dataset_proto.update_time.value if dataset_proto.HasField('update_time') else None
        catalog_name = dataset_proto.catalog_name.value if dataset_proto.HasField('catalog_name') else None
        catalog_type = dataset_proto.catalog_type.value if dataset_proto.HasField('catalog_type') else None
        catalog_database = dataset_proto.catalog_database.value if dataset_proto.HasField('catalog_database') else None
        catalog_connection_uri = dataset_proto.catalog_connection_uri.value \
            if dataset_proto.HasField('catalog_connection_uri') else None
        catalog_table = dataset_proto.catalog_table.value if dataset_proto.HasField('catalog_table') else None
        schema = Schema(name_list=name_list, type_list=data_type_list)
        return DatasetMeta(uuid=dataset_proto.uuid,
                           name=dataset_proto.name,
                           data_format=data_format,
                           description=description,
                           uri=uri,
                           create_time=create_time,
                           update_time=update_time,
                           properties=properties,
                           schema=schema,
                           catalog_name=catalog_name,
                           catalog_type=catalog_type,
                           catalog_database=catalog_database,
                           catalog_connection_uri=catalog_connection_uri,
                           catalog_table=catalog_table)

    @staticmethod
    def proto_to_dataset_meta_list(dataset_proto_list: List[DatasetProto]) -> List[DatasetMeta]:
        list_dataset_meta = []
        for dataset_proto in dataset_proto_list:
            list_dataset_meta.append(ProtoToMeta.proto_to_dataset_meta(dataset_proto))
        return list_dataset_meta

    @staticmethod
    def proto_to_project_meta(project_proto: ProjectProto) -> ProjectMeta:
        properties = project_proto.properties
        if properties == {}:
            properties = None
        return ProjectMeta(
            uuid=project_proto.uuid,
            name=project_proto.name,
            properties=properties,
            uri=project_proto.uri.value if project_proto.HasField('uri') else None)

    @staticmethod
    def proto_to_project_meta_list(projects: List[ProjectProto]) -> List[ProjectMeta]:
        project_meta_list = []
        for project in projects:
            project_meta_list.append(ProtoToMeta.proto_to_project_meta(project))
        return project_meta_list

    @staticmethod
    def proto_to_workflow_meta(workflow_proto: WorkflowMetaProto) -> WorkflowMeta:
        properties = workflow_proto.properties
        if properties == {}:
            properties = None
        project_id = workflow_proto.project_id.value if workflow_proto.HasField('project_id') else None
        create_time = workflow_proto.create_time.value if workflow_proto.HasField('create_time') else None
        update_time = workflow_proto.update_time.value if workflow_proto.HasField('update_time') else None
        return WorkflowMeta(uuid=workflow_proto.uuid,
                            name=workflow_proto.name,
                            project_id=project_id,
                            properties=properties,
                            create_time=create_time,
                            update_time=update_time)

    @staticmethod
    def proto_to_workflow_meta_list(workflows: List[WorkflowMetaProto]) -> List[WorkflowMeta]:
        workflow_meta_list = []
        for workflow in workflows:
            workflow_meta_list.append(ProtoToMeta.proto_to_workflow_meta(workflow))
        return workflow_meta_list

    @staticmethod
    def proto_to_artifact_meta(artifact_proto: ArtifactProto) -> ArtifactMeta:
        properties = artifact_proto.properties
        if properties == {}:
            properties = None
        return ArtifactMeta(
            uuid=artifact_proto.uuid,
            name=artifact_proto.name,
            properties=properties,
            artifact_type=artifact_proto.artifact_type.value if artifact_proto.HasField('artifact_type') else None,
            description=artifact_proto.description.value if artifact_proto.HasField('description') else None,
            uri=artifact_proto.uri.value if artifact_proto.HasField('uri') else None,
            create_time=artifact_proto.create_time.value if artifact_proto.HasField('create_time') else None,
            update_time=artifact_proto.update_time.value if artifact_proto.HasField('update_time') else None)

    @staticmethod
    def proto_to_artifact_meta_list(artifacts: List[ArtifactProto]) -> List[ArtifactMeta]:
        artifact_meta_list = []
        for artifact in artifacts:
            artifact_meta_list.append(ProtoToMeta.proto_to_artifact_meta(artifact))
        return artifact_meta_list

    @staticmethod
    def proto_to_state(state):
        if state == StateProto.INIT:
            return State.INIT
        elif state == StateProto.STARTING:
            return State.STARTING
        elif state == StateProto.RUNNING:
            return State.RUNNING
        elif state == StateProto.FINISHED:
            return State.FINISHED
        elif state == StateProto.FAILED:
            return State.FAILED
        elif state == StateProto.KILLING:
            return State.KILLING
        elif state == StateProto.KILLED:
            return State.KILLED

    @staticmethod
    def proto_to_model_relation_meta(model_relation_proto: ModelRelationProto) -> ModelRelationMeta:
        return ModelRelationMeta(uuid=model_relation_proto.uuid,
                                 name=model_relation_proto.name,
                                 project_id=model_relation_proto.project_id.value
                                 if model_relation_proto.HasField('project_id') else None)

    @staticmethod
    def proto_to_model_relation_meta_list(model_proto_relation_list: List[ModelRelationProto]) -> List[
        ModelRelationMeta]:
        model_relation_meta_list = []
        for model_proto in model_proto_relation_list:
            model_relation_meta_list.append(ProtoToMeta.proto_to_model_relation_meta(model_proto))
        return model_relation_meta_list

    @staticmethod
    def proto_to_model_meta(model_proto: ModelProto) -> ModelMeta:
        return ModelMeta(uuid=model_proto.uuid, name=model_proto.name,
                         model_desc=model_proto.model_desc.value if model_proto.HasField(
                             'model_desc') else None,
                         project_id=model_proto.project_id.value if model_proto.HasField('project_id') else None)

    @staticmethod
    def proto_to_model_meta_list(model_proto_list: List[ModelProto]) -> List[ModelMeta]:
        model_meta_list = []
        for model_proto in model_proto_list:
            model_meta_list.append(ProtoToMeta.proto_to_model_meta(model_proto))
        return model_meta_list

    @staticmethod
    def proto_to_model_version_relation_meta(
            model_version_proto: ModelVersionRelationProto) -> ModelVersionRelationMeta:
        return ModelVersionRelationMeta(
            version=model_version_proto.version.value if model_version_proto.HasField('version') else None,
            model_id=model_version_proto.model_id.value if model_version_proto.HasField('model_id') else None,
            project_snapshot_id=model_version_proto.project_snapshot_id.value
            if model_version_proto.HasField('project_snapshot_id') else None)

    @staticmethod
    def proto_to_model_version_relation_meta_list(model_version_proto_list: List[ModelVersionRelationProto]) -> List[
        ModelVersionRelationMeta]:
        model_version_relation_meta_list = []
        for model_version_proto in model_version_proto_list:
            model_version_relation_meta_list.append(
                ProtoToMeta.proto_to_model_version_relation_meta(model_version_proto))
        return model_version_relation_meta_list

    @staticmethod
    def proto_to_model_version_meta(
            model_version_proto: ModelVersionProto) -> ModelVersionMeta:
        return ModelVersionMeta(
            version=model_version_proto.version.value if model_version_proto.HasField(
                'version') else None,
            model_id=model_version_proto.model_id.value if model_version_proto.HasField('model_id') else None,
            model_path=model_version_proto.model_path.value if model_version_proto.HasField(
                'model_path') else None,
            model_type=model_version_proto.model_type.value if model_version_proto.HasField(
                'model_type') else None,
            project_snapshot_id=model_version_proto.project_snapshot_id.value if model_version_proto.HasField(
                'project_snapshot_id') else None,
            version_desc=model_version_proto.version_desc.value if model_version_proto.HasField(
                'version_desc') else None,
            current_stage=ModelVersionStage.Name(model_version_proto.current_stage))

    @staticmethod
    def proto_to_model_version_meta_list(model_version_proto_list: List[ModelVersionProto]) -> List[ModelVersionMeta]:
        model_version_meta_list = []
        for model_version_proto in model_version_proto_list:
            model_version_meta_list.append(
                ProtoToMeta.proto_to_model_version_meta(model_version_proto))
        return model_version_meta_list



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import unittest

import sqlalchemy

from ai_flow.client.ai_flow_client import AIFlowClient
from ai_flow.endpoint.server.server import AIFlowServer
from ai_flow.store.db.base_model import base
from ai_flow.test.endpoint import test_client
from ai_flow.test.store.test_sqlalchemy_store import _get_store
from ai_flow.test.test_util import get_mysql_server_url

_PORT = '50051'


@unittest.skip("To run this test you need to configure the mysql info in 'ai_flow/test/test_util.py'")
class TestAIFlowClientMySQL(test_client.TestAIFlowClientSqlite):

    @classmethod
    def setUpClass(cls) -> None:
        print("TestAIFlowClientMySQL setUpClass")
        db_server_url = get_mysql_server_url()
        cls.db_name = 'test_aiflow_client'
        cls.engine = sqlalchemy.create_engine(db_server_url)
        cls.engine.execute('DROP DATABASE IF EXISTS %s' % cls.db_name)
        cls.engine.execute('CREATE DATABASE IF NOT EXISTS %s' % cls.db_name)
        cls.store_uri = '%s/%s' % (db_server_url, cls.db_name)
        cls.server = AIFlowServer(store_uri=cls.store_uri, port=_PORT)
        cls.server.run()
        test_client.client = AIFlowClient(server_uri='localhost:' + _PORT)
        test_client.client1 = AIFlowClient(server_uri='localhost:' + _PORT)
        test_client.client2 = AIFlowClient(server_uri='localhost:' + _PORT)

    @classmethod
    def tearDownClass(cls) -> None:
        test_client.client.stop_listen_event()
        test_client.client1.stop_listen_event()
        test_client.client2.stop_listen_event()
        cls.server.stop()

    def setUp(self) -> None:
        _get_store(self.store_uri)

    def tearDown(self) -> None:
        store = _get_store(self.store_uri)
        base.metadata.drop_all(store.db_engine)


if __name__ == '__main__':
    unittest.main()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import datetime


def max_partition(
    table, schema="default", field=None, filter_map=None, metastore_conn_id='metastore_default'
):
    """
    Gets the max partition for a table.

    :param schema: The hive schema the table lives in
    :type schema: str
    :param table: The hive table you are interested in, supports the dot
        notation as in "my_database.my_table", if a dot is found,
        the schema param is disregarded
    :type table: str
    :param metastore_conn_id: The hive connection you are interested in.
        If your default is set you don't need to use this parameter.
    :type metastore_conn_id: str
    :param filter_map: partition_key:partition_value map used for partition filtering,
                       e.g. {'key1': 'value1', 'key2': 'value2'}.
                       Only partitions matching all partition_key:partition_value
                       pairs will be considered as candidates of max partition.
    :type filter_map: dict
    :param field: the field to get the max value from. If there's only
        one partition field, this will be inferred
    :type field: str

    >>> max_partition('airflow.static_babynames_partitioned')
    '2015-01-01'
    """
    from airflow.providers.apache.hive.hooks.hive import HiveMetastoreHook

    if '.' in table:
        schema, table = table.split('.')
    hive_hook = HiveMetastoreHook(metastore_conn_id=metastore_conn_id)
    return hive_hook.max_partition(schema=schema, table_name=table, field=field, filter_map=filter_map)


def _closest_date(target_dt, date_list, before_target=None):
    """
    This function finds the date in a list closest to the target date.
    An optional parameter can be given to get the closest before or after.

    :param target_dt: The target date
    :type target_dt: datetime.date
    :param date_list: The list of dates to search
    :type date_list: list[datetime.date]
    :param before_target: closest before or after the target
    :type before_target: bool or None
    :returns: The closest date
    :rtype: datetime.date or None
    """
    time_before = lambda d: target_dt - d if d <= target_dt else datetime.timedelta.max
    time_after = lambda d: d - target_dt if d >= target_dt else datetime.timedelta.max
    any_time = lambda d: target_dt - d if d < target_dt else d - target_dt
    if before_target is None:
        return min(date_list, key=any_time).date()
    if before_target:
        return min(date_list, key=time_before).date()
    else:
        return min(date_list, key=time_after).date()


def closest_ds_partition(table, ds, before=True, schema="default", metastore_conn_id='metastore_default'):
    """
    This function finds the date in a list closest to the target date.
    An optional parameter can be given to get the closest before or after.

    :param table: A hive table name
    :type table: str
    :param ds: A datestamp ``%Y-%m-%d`` e.g. ``yyyy-mm-dd``
    :type ds: list[datetime.date]
    :param before: closest before (True), after (False) or either side of ds
    :type before: bool or None
    :param schema: table schema
    :type schema: str
    :param metastore_conn_id: which metastore connection to use
    :type metastore_conn_id: str
    :returns: The closest date
    :rtype: str or None

    >>> tbl = 'airflow.static_babynames_partitioned'
    >>> closest_ds_partition(tbl, '2015-01-02')
    '2015-01-01'
    """
    from airflow.providers.apache.hive.hooks.hive import HiveMetastoreHook

    if '.' in table:
        schema, table = table.split('.')
    hive_hook = HiveMetastoreHook(metastore_conn_id=metastore_conn_id)
    partitions = hive_hook.get_partitions(schema=schema, table_name=table)
    if not partitions:
        return None
    part_vals = [list(p.values())[0] for p in partitions]
    if ds in part_vals:
        return ds
    else:
        parts = [datetime.datetime.strptime(pv, '%Y-%m-%d') for pv in part_vals]
        target_dt = datetime.datetime.strptime(ds, '%Y-%m-%d')
        closest_ds = _closest_date(target_dt, parts, before_target=before)
        return closest_ds.isoformat()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import message_pb2 as message__pb2
from . import model_center_service_pb2 as model__center__service__pb2


class ModelCenterServiceStub(object):
    """AIFlowService provides model registry service rest endpoint of ModelCenterService for Model Center component.
    Functions of ModelCenterService include:
    1.Create registered model
    2.Update registered model
    3.Delete registered model
    4.List registered models
    5.Get registered model detail
    6.Create model version
    7.Update model version
    8.Delete model version
    9.Get model version detail
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.createRegisteredModel = channel.unary_unary(
                '/ai_flow.ModelCenterService/createRegisteredModel',
                request_serializer=model__center__service__pb2.CreateRegisteredModelRequest.SerializeToString,
                response_deserializer=message__pb2.Response.FromString,
                )
        self.updateRegisteredModel = channel.unary_unary(
                '/ai_flow.ModelCenterService/updateRegisteredModel',
                request_serializer=model__center__service__pb2.UpdateRegisteredModelRequest.SerializeToString,
                response_deserializer=message__pb2.Response.FromString,
                )
        self.deleteRegisteredModel = channel.unary_unary(
                '/ai_flow.ModelCenterService/deleteRegisteredModel',
                request_serializer=model__center__service__pb2.DeleteRegisteredModelRequest.SerializeToString,
                response_deserializer=message__pb2.Response.FromString,
                )
        self.listRegisteredModels = channel.unary_unary(
                '/ai_flow.ModelCenterService/listRegisteredModels',
                request_serializer=model__center__service__pb2.ListRegisteredModelsRequest.SerializeToString,
                response_deserializer=message__pb2.Response.FromString,
                )
        self.getRegisteredModelDetail = channel.unary_unary(
                '/ai_flow.ModelCenterService/getRegisteredModelDetail',
                request_serializer=model__center__service__pb2.GetRegisteredModelDetailRequest.SerializeToString,
                response_deserializer=message__pb2.Response.FromString,
                )
        self.createModelVersion = channel.unary_unary(
                '/ai_flow.ModelCenterService/createModelVersion',
                request_serializer=model__center__service__pb2.CreateModelVersionRequest.SerializeToString,
                response_deserializer=message__pb2.Response.FromString,
                )
        self.updateModelVersion = channel.unary_unary(
                '/ai_flow.ModelCenterService/updateModelVersion',
                request_serializer=model__center__service__pb2.UpdateModelVersionRequest.SerializeToString,
                response_deserializer=message__pb2.Response.FromString,
                )
        self.deleteModelVersion = channel.unary_unary(
                '/ai_flow.ModelCenterService/deleteModelVersion',
                request_serializer=model__center__service__pb2.DeleteModelVersionRequest.SerializeToString,
                response_deserializer=message__pb2.Response.FromString,
                )
        self.getModelVersionDetail = channel.unary_unary(
                '/ai_flow.ModelCenterService/getModelVersionDetail',
                request_serializer=model__center__service__pb2.GetModelVersionDetailRequest.SerializeToString,
                response_deserializer=message__pb2.Response.FromString,
                )


class ModelCenterServiceServicer(object):
    """AIFlowService provides model registry service rest endpoint of ModelCenterService for Model Center component.
    Functions of ModelCenterService include:
    1.Create registered model
    2.Update registered model
    3.Delete registered model
    4.List registered models
    5.Get registered model detail
    6.Create model version
    7.Update model version
    8.Delete model version
    9.Get model version detail
    """

    def createRegisteredModel(self, request, context):
        """Create registered model with metadata of RegisteredModel.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def updateRegisteredModel(self, request, context):
        """Update registered model with metadata of RegisteredModel.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def deleteRegisteredModel(self, request, context):
        """Delete registered model with metadata of RegisteredModel.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def listRegisteredModels(self, request, context):
        """List registered models about metadata of RegisteredModel.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getRegisteredModelDetail(self, request, context):
        """Get registered model detail including metadata of RegisteredModel.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def createModelVersion(self, request, context):
        """Create model version with metadata of ModelVersion.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def updateModelVersion(self, request, context):
        """Update model version with metadata of ModelVersion.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def deleteModelVersion(self, request, context):
        """Delete model version with metadata of ModelVersion.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getModelVersionDetail(self, request, context):
        """Get model version detail with metadata of ModelVersion.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ModelCenterServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'createRegisteredModel': grpc.unary_unary_rpc_method_handler(
                    servicer.createRegisteredModel,
                    request_deserializer=model__center__service__pb2.CreateRegisteredModelRequest.FromString,
                    response_serializer=message__pb2.Response.SerializeToString,
            ),
            'updateRegisteredModel': grpc.unary_unary_rpc_method_handler(
                    servicer.updateRegisteredModel,
                    request_deserializer=model__center__service__pb2.UpdateRegisteredModelRequest.FromString,
                    response_serializer=message__pb2.Response.SerializeToString,
            ),
            'deleteRegisteredModel': grpc.unary_unary_rpc_method_handler(
                    servicer.deleteRegisteredModel,
                    request_deserializer=model__center__service__pb2.DeleteRegisteredModelRequest.FromString,
                    response_serializer=message__pb2.Response.SerializeToString,
            ),
            'listRegisteredModels': grpc.unary_unary_rpc_method_handler(
                    servicer.listRegisteredModels,
                    request_deserializer=model__center__service__pb2.ListRegisteredModelsRequest.FromString,
                    response_serializer=message__pb2.Response.SerializeToString,
            ),
            'getRegisteredModelDetail': grpc.unary_unary_rpc_method_handler(
                    servicer.getRegisteredModelDetail,
                    request_deserializer=model__center__service__pb2.GetRegisteredModelDetailRequest.FromString,
                    response_serializer=message__pb2.Response.SerializeToString,
            ),
            'createModelVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.createModelVersion,
                    request_deserializer=model__center__service__pb2.CreateModelVersionRequest.FromString,
                    response_serializer=message__pb2.Response.SerializeToString,
            ),
            'updateModelVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.updateModelVersion,
                    request_deserializer=model__center__service__pb2.UpdateModelVersionRequest.FromString,
                    response_serializer=message__pb2.Response.SerializeToString,
            ),
            'deleteModelVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.deleteModelVersion,
                    request_deserializer=model__center__service__pb2.DeleteModelVersionRequest.FromString,
                    response_serializer=message__pb2.Response.SerializeToString,
            ),
            'getModelVersionDetail': grpc.unary_unary_rpc_method_handler(
                    servicer.getModelVersionDetail,
                    request_deserializer=model__center__service__pb2.GetModelVersionDetailRequest.FromString,
                    response_serializer=message__pb2.Response.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ai_flow.ModelCenterService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ModelCenterService(object):
    """AIFlowService provides model registry service rest endpoint of ModelCenterService for Model Center component.
    Functions of ModelCenterService include:
    1.Create registered model
    2.Update registered model
    3.Delete registered model
    4.List registered models
    5.Get registered model detail
    6.Create model version
    7.Update model version
    8.Delete model version
    9.Get model version detail
    """

    @staticmethod
    def createRegisteredModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.ModelCenterService/createRegisteredModel',
            model__center__service__pb2.CreateRegisteredModelRequest.SerializeToString,
            message__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def updateRegisteredModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.ModelCenterService/updateRegisteredModel',
            model__center__service__pb2.UpdateRegisteredModelRequest.SerializeToString,
            message__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def deleteRegisteredModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.ModelCenterService/deleteRegisteredModel',
            model__center__service__pb2.DeleteRegisteredModelRequest.SerializeToString,
            message__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def listRegisteredModels(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.ModelCenterService/listRegisteredModels',
            model__center__service__pb2.ListRegisteredModelsRequest.SerializeToString,
            message__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getRegisteredModelDetail(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.ModelCenterService/getRegisteredModelDetail',
            model__center__service__pb2.GetRegisteredModelDetailRequest.SerializeToString,
            message__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def createModelVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.ModelCenterService/createModelVersion',
            model__center__service__pb2.CreateModelVersionRequest.SerializeToString,
            message__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def updateModelVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.ModelCenterService/updateModelVersion',
            model__center__service__pb2.UpdateModelVersionRequest.SerializeToString,
            message__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def deleteModelVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.ModelCenterService/deleteModelVersion',
            model__center__service__pb2.DeleteModelVersionRequest.SerializeToString,
            message__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getModelVersionDetail(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ai_flow.ModelCenterService/getModelVersionDetail',
            model__center__service__pb2.GetModelVersionDetailRequest.SerializeToString,
            message__pb2.Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Manages all providers."""
import fnmatch
import importlib
import json
import logging
import os
from collections import OrderedDict
from typing import Any, Dict, NamedTuple, Set

import jsonschema
import yaml
from wtforms import Field

from airflow.utils.entry_points import entry_points_with_dist

try:
    import importlib.resources as importlib_resources
except ImportError:
    # Try back-ported to PY<37 `importlib_resources`.
    import importlib_resources

log = logging.getLogger(__name__)


def _create_provider_schema_validator():
    """Creates JSON schema validator from the provider.yaml.schema.json"""
    schema = json.loads(importlib_resources.read_text('airflow', 'provider.yaml.schema.json'))
    cls = jsonschema.validators.validator_for(schema)
    validator = cls(schema)
    return validator


def _create_customized_form_field_behaviours_schema_validator():
    """Creates JSON schema validator from the customized_form_field_behaviours.schema.json"""
    schema = json.loads(
        importlib_resources.read_text('airflow', 'customized_form_field_behaviours.schema.json')
    )
    cls = jsonschema.validators.validator_for(schema)
    validator = cls(schema)
    return validator


class ProviderInfo(NamedTuple):
    """Provider information"""

    version: str
    provider_info: Dict


class HookInfo(NamedTuple):
    """Hook information"""

    connection_class: str
    connection_id_attribute_name: str
    package_name: str
    hook_name: str


class ConnectionFormWidgetInfo(NamedTuple):
    """Connection Form Widget information"""

    connection_class: str
    package_name: str
    field: Field


class ProvidersManager:
    """
    Manages all provider packages. This is a Singleton class. The first time it is
    instantiated, it discovers all available providers in installed packages and
    local source folders (if airflow is run from sources).
    """

    _instance = None
    resource_version = "0"

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        # Keeps dict of providers keyed by module name
        self._provider_dict: Dict[str, ProviderInfo] = {}
        # Keeps dict of hooks keyed by connection type
        self._hooks_dict: Dict[str, HookInfo] = {}
        # Keeps methods that should be used to add custom widgets tuple of keyed by name of the extra field
        self._connection_form_widgets: Dict[str, ConnectionFormWidgetInfo] = {}
        # Customizations for javascript fields are kept here
        self._field_behaviours: Dict[str, Dict] = {}
        self._extra_link_class_name_set: Set[str] = set()
        self._provider_schema_validator = _create_provider_schema_validator()
        self._customized_form_fields_schema_validator = (
            _create_customized_form_field_behaviours_schema_validator()
        )
        self._initialized = False

    def initialize_providers_manager(self):
        """Lazy initialization of provider data."""
        # We cannot use @cache here because it does not work during pytests, apparently each test
        # runs it it's own namespace and ProvidersManager is a different object in each namespace
        # even if it is singleton but @cache on the initialize_providers_manager message still works in the
        # way that it is called only once for one of the objects (at least this is how it looks like
        # from running tests)
        if self._initialized:
            return
        # Local source folders are loaded first. They should take precedence over the package ones for
        # Development purpose. In production provider.yaml files are not present in the 'airflow" directory
        # So there is no risk we are going to override package provider accidentally. This can only happen
        # in case of local development
        self._discover_all_airflow_builtin_providers_from_local_sources()
        self._discover_all_providers_from_packages()
        self._discover_hooks()
        self._provider_dict = OrderedDict(sorted(self._provider_dict.items()))  # noqa
        self._hooks_dict = OrderedDict(sorted(self._hooks_dict.items()))  # noqa
        self._connection_form_widgets = OrderedDict(sorted(self._connection_form_widgets.items()))  # noqa
        self._field_behaviours = OrderedDict(sorted(self._field_behaviours.items()))  # noqa
        self._discover_extra_links()
        self._initialized = True

    def _discover_all_providers_from_packages(self) -> None:
        """
        Discovers all providers by scanning packages installed. The list of providers should be returned
        via the 'apache_airflow_provider' entrypoint as a dictionary conforming to the
        'airflow/provider.yaml.schema.json' schema.
        """
        for entry_point, dist in entry_points_with_dist('apache_airflow_provider'):
            package_name = dist.metadata['name']
            log.debug("Loading %s from package %s", entry_point, package_name)
            version = dist.version
            provider_info = entry_point.load()()
            self._provider_schema_validator.validate(provider_info)
            provider_info_package_name = provider_info['package-name']
            if package_name != provider_info_package_name:
                raise Exception(
                    f"The package '{package_name}' from setuptools and "
                    f"{provider_info_package_name} do not match. Please make sure they are aligned"
                )
            if package_name not in self._provider_dict:
                self._provider_dict[package_name] = (version, provider_info)
            else:
                log.warning(
                    "The provider for package '%s' could not be registered from because providers for that "
                    "package name have already been registered",
                    package_name,
                )

    def _discover_all_airflow_builtin_providers_from_local_sources(self) -> None:
        """
        Finds all built-in airflow providers if airflow is run from the local sources.
        It finds `provider.yaml` files for all such providers and registers the providers using those.

        This 'provider.yaml' scanning takes precedence over scanning packages installed
        in case you have both sources and packages installed, the providers will be loaded from
        the "airflow" sources rather than from the packages.
        """
        try:
            import airflow.providers
        except ImportError:
            log.info("You have no providers installed.")
            return
        try:
            for path in airflow.providers.__path__:
                self._add_provider_info_from_local_source_files_on_path(path)
        except Exception as e:  # noqa pylint: disable=broad-except
            log.warning("Error when loading 'provider.yaml' files from airflow sources: %s", e)

    def _add_provider_info_from_local_source_files_on_path(self, path) -> None:
        """
        Finds all the provider.yaml files in the directory specified.

        :param path: path where to look for provider.yaml files
        """
        root_path = path
        for folder, subdirs, files in os.walk(path, topdown=True):
            for filename in fnmatch.filter(files, "provider.yaml"):
                package_name = "apache-airflow-providers" + folder[len(root_path) :].replace(os.sep, "-")
                self._add_provider_info_from_local_source_file(os.path.join(folder, filename), package_name)
                subdirs[:] = []

    def _add_provider_info_from_local_source_file(self, path, package_name) -> None:
        """
        Parses found provider.yaml file and adds found provider to the dictionary.

        :param path: full file path of the provider.yaml file
        :param package_name: name of the package
        """
        try:
            log.debug("Loading %s from %s", package_name, path)
            with open(path) as provider_yaml_file:
                provider_info = yaml.safe_load(provider_yaml_file)
            self._provider_schema_validator.validate(provider_info)

            version = provider_info['versions'][0]
            if package_name not in self._provider_dict:
                self._provider_dict[package_name] = (version, provider_info)
            else:
                log.warning(
                    "The providers for package '%s' could not be registered because providers for that "
                    "package name have already been registered",
                    package_name,
                )
        except Exception as e:  # noqa pylint: disable=broad-except
            log.warning("Error when loading '%s': %s", path, e)

    def _discover_hooks(self) -> None:
        """Retrieves all connections defined in the providers"""
        for name, provider in self._provider_dict.items():
            provider_package = name
            hook_class_names = provider[1].get("hook-class-names")
            if hook_class_names:
                for hook_class_name in hook_class_names:
                    self._add_hook(hook_class_name, provider_package)

    @staticmethod
    def _get_attr(obj: Any, attr_name: str):
        """Retrieves attributes of an object, or warns if not found"""
        if not hasattr(obj, attr_name):
            log.warning("The '%s' is missing %s attribute and cannot be registered", obj, attr_name)
            return None
        return getattr(obj, attr_name)

    def _add_hook(self, hook_class_name: str, provider_package: str) -> None:
        """
        Adds hook class name to list of hooks

        :param hook_class_name: name of the Hook class
        :param provider_package: provider package adding the hook
        """
        if provider_package.startswith("apache-airflow"):
            provider_path = provider_package[len("apache-") :].replace("-", ".")
            if not hook_class_name.startswith(provider_path):
                log.warning(
                    "Sanity check failed when importing '%s' from '%s' package. It should start with '%s'",
                    hook_class_name,
                    provider_package,
                    provider_path,
                )
                return
        if hook_class_name in self._hooks_dict:
            log.warning(
                "The hook_class '%s' has been already registered.",
                hook_class_name,
            )
            return
        try:
            module, class_name = hook_class_name.rsplit('.', maxsplit=1)
            hook_class = getattr(importlib.import_module(module), class_name)
            # Do not use attr here. We want to check only direct class fields not those
            # inherited from parent hook. This way we add form fields only once for the whole
            # hierarchy and we add it only from the parent hook that provides those!
            if 'get_connection_form_widgets' in hook_class.__dict__:
                widgets = hook_class.get_connection_form_widgets()
                if widgets:
                    self._add_widgets(provider_package, hook_class, widgets)
            if 'get_ui_field_behaviour' in hook_class.__dict__:
                field_behaviours = hook_class.get_ui_field_behaviour()
                if field_behaviours:
                    self._add_customized_fields(provider_package, hook_class, field_behaviours)

        except Exception as e:  # noqa pylint: disable=broad-except
            log.warning(
                "Exception when importing '%s' from '%s' package: %s",
                hook_class_name,
                provider_package,
                e,
            )
            return

        conn_type: str = self._get_attr(hook_class, 'conn_type')
        connection_id_attribute_name: str = self._get_attr(hook_class, 'conn_name_attr')
        hook_name: str = self._get_attr(hook_class, 'hook_name')

        if not conn_type or not connection_id_attribute_name or not hook_name:
            return

        self._hooks_dict[conn_type] = HookInfo(
            hook_class_name,
            connection_id_attribute_name,
            provider_package,
            hook_name,
        )

    def _add_widgets(self, package_name: str, hook_class: type, widgets: Dict[str, Field]):
        for field_name, field in widgets.items():
            if not field_name.startswith("extra__"):
                log.warning(
                    "The field %s from class %s does not start with 'extra__'. Ignoring it.",
                    field_name,
                    hook_class.__name__,
                )
                continue
            if field_name in self._connection_form_widgets:
                log.warning(
                    "The field %s from class %s has already been added by another provider. Ignoring it.",
                    field_name,
                    hook_class.__name__,
                )
                # In case of inherited hooks this might be happening several times
                continue
            self._connection_form_widgets[field_name] = ConnectionFormWidgetInfo(
                hook_class.__name__, package_name, field
            )

    def _add_customized_fields(self, package_name: str, hook_class: type, customized_fields: Dict):
        try:
            connection_type = getattr(hook_class, "conn_type")
            self._customized_form_fields_schema_validator.validate(customized_fields)
            if connection_type in self._field_behaviours:
                log.warning(
                    "The connection_type %s from package %s and class %s has already been added "
                    "by another provider. Ignoring it.",
                    connection_type,
                    package_name,
                    hook_class.__name__,
                )
                return
            self._field_behaviours[connection_type] = customized_fields
        except Exception as e:  # noqa pylint: disable=broad-except
            log.warning(
                "Error when loading customized fields from package '%s' hook class '%s': %s",
                package_name,
                hook_class.__name__,
                e,
            )

    def _discover_extra_links(self) -> None:
        """Retrieves all extra links defined in the providers"""
        for provider_package, (_, provider) in self._provider_dict.items():
            if provider.get("extra-links"):
                for extra_link in provider["extra-links"]:
                    self._add_extra_link(extra_link, provider_package)

    def _add_extra_link(self, extra_link_class_name, provider_package) -> None:
        """
        Adds extra link class name to the list of classes
        :param extra_link_class_name: name of the class to add
        :param provider_package: provider package adding the link
        :return:
        """
        if provider_package.startswith("apache-airflow"):
            provider_path = provider_package[len("apache-") :].replace("-", ".")
            if not extra_link_class_name.startswith(provider_path):
                log.warning(
                    "Sanity check failed when importing '%s' from '%s' package. It should start with '%s'",
                    extra_link_class_name,
                    provider_package,
                    provider_path,
                )
                return
        self._extra_link_class_name_set.add(extra_link_class_name)

    @property
    def providers(self) -> Dict[str, ProviderInfo]:
        """Returns information about available providers."""
        self.initialize_providers_manager()
        return self._provider_dict

    @property
    def hooks(self) -> Dict[str, HookInfo]:
        """Returns dictionary of connection_type-to-hook mapping"""
        self.initialize_providers_manager()
        return self._hooks_dict

    @property
    def extra_links_class_names(self):
        """Returns set of extra link class names."""
        self.initialize_providers_manager()
        return sorted(self._extra_link_class_name_set)

    @property
    def connection_form_widgets(self) -> Dict[str, ConnectionFormWidgetInfo]:
        """Returns widgets for connection forms."""
        self.initialize_providers_manager()
        return self._connection_form_widgets

    @property
    def field_behaviours(self) -> Dict[str, Dict]:
        """Returns dictionary with field behaviours for connection types."""
        self.initialize_providers_manager()
        return self._field_behaviours



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from functools import wraps

from ai_flow.meta.workflow_meta import WorkflowMeta
from google.protobuf.json_format import MessageToJson, Parse

from ai_flow.common.status import Status
from ai_flow.meta.artifact_meta import ArtifactMeta
from ai_flow.meta.dataset_meta import DataType, Schema, DatasetMeta
from ai_flow.meta.model_meta import ModelMeta, ModelVersionMeta
from ai_flow.meta.model_relation_meta import ModelRelationMeta, ModelVersionRelationMeta
from ai_flow.meta.project_meta import ProjectMeta
from ai_flow.metadata_store.utils.MetaToProto import MetaToProto
from ai_flow.metadata_store.utils.ProtoToMeta import ProtoToMeta
from ai_flow.protobuf.message_pb2 import Response, SUCCESS, ReturnCode, RESOURCE_DOES_NOT_EXIST, \
    DatasetProto, ModelProto, ModelVersionProto, ProjectProto, INTERNAL_ERROR, \
    DataTypeProto, ArtifactProto, WorkflowMetaProto
from ai_flow.protobuf.metadata_service_pb2 import DatasetListProto, \
    ProjectListProto, ModelVersionRelationListProto, ModelRelationListProto, ModelVersionListProto, \
    ArtifactListProto, WorkflowListProto
from ai_flow.endpoint.server.exception import AIFlowException
from ai_flow.store.sqlalchemy_store import UPDATE_FAIL


def catch_exception(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except AIFlowException as e:
            return Response(return_code=str(e.error_code), return_msg=e.error_msg)

    return wrapper


def _wrap_response(response_message):
    if response_message is None:
        return Response(return_code=str(SUCCESS), return_msg=ReturnCode.Name(SUCCESS).lower())
    else:
        return Response(return_code=str(SUCCESS), return_msg=ReturnCode.Name(SUCCESS).lower(),
                        data=MessageToJson(response_message, preserving_proto_field_name=True))


def _parse_response(response, message):
    if response.return_code == str(SUCCESS):
        if response.data == '':
            return None
        else:
            return Parse(response.data, message, ignore_unknown_fields=False)
    else:
        raise AIFlowException(error_code=response.return_code, error_msg=response.return_msg)


def _unwrap_dataset_response(response):
    if response.return_code == str(SUCCESS):
        return ProtoToMeta.proto_to_dataset_meta(Parse(response.data, DatasetProto()))
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_dataset_list_response(response):
    if response.return_code == str(SUCCESS):
        dataset_proto_list = Parse(response.data, DatasetListProto())
        return ProtoToMeta.proto_to_dataset_meta_list(dataset_proto_list.datasets)
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_model_relation_response(response):
    if response.return_code == str(SUCCESS):
        return ProtoToMeta.proto_to_model_relation_meta(Parse(response.data, ModelProto()))
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_model_relation_list_response(response):
    if response.return_code == str(SUCCESS):
        model_proto_list = Parse(response.data, ModelRelationListProto())
        return ProtoToMeta.proto_to_model_relation_meta_list(model_proto_list.model_relations)
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_model_version_relation_response(response):
    if response.return_code == str(SUCCESS):
        return ProtoToMeta.proto_to_model_version_relation_meta(Parse(response.data, ModelVersionProto()))
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_model_version_relation_list_response(response):
    if response.return_code == str(SUCCESS):
        model_version_proto_list = Parse(response.data, ModelVersionRelationListProto())
        return ProtoToMeta.proto_to_model_version_relation_meta_list(model_version_proto_list.model_versions)
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_model_response(response):
    if response.return_code == str(SUCCESS):
        return ProtoToMeta.proto_to_model_meta(Parse(response.data, ModelProto()))
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_model_version_response(response):
    if response.return_code == str(SUCCESS):
        return ProtoToMeta.proto_to_model_version_meta(Parse(response.data, ModelVersionProto()))
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_model_version_list_response(response):
    if response.return_code == str(SUCCESS):
        model_version_proto_list = Parse(response.data, ModelVersionListProto())
        return ProtoToMeta.proto_to_model_version_meta_list(model_version_proto_list.model_versions)
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_project_response(response):
    if response.return_code == str(SUCCESS):
        return ProtoToMeta.proto_to_project_meta(Parse(response.data, ProjectProto()))
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_project_list_response(response):
    if response.return_code == str(SUCCESS):
        project_proto_list = Parse(response.data, ProjectListProto())
        return ProtoToMeta.proto_to_project_meta_list(project_proto_list.projects)
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_workflow_response(response):
    if response.return_code == str(SUCCESS):
        return ProtoToMeta.proto_to_workflow_meta(Parse(response.data, WorkflowMetaProto()))
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_workflow_list_response(response):
    if response.return_code == str(SUCCESS):
        workflow_proto_list = Parse(response.data, WorkflowListProto())
        return ProtoToMeta.proto_to_workflow_meta_list(workflow_proto_list.workflows)
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_artifact_response(response):
    if response.return_code == str(SUCCESS):
        return ProtoToMeta.proto_to_artifact_meta(Parse(response.data, ArtifactProto()))
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_artifact_list_response(response):
    if response.return_code == str(SUCCESS):
        artifact_proto_list = Parse(response.data, ArtifactListProto())
        return ProtoToMeta.proto_to_artifact_meta_list(artifact_proto_list.artifacts)
    elif response.return_code == str(RESOURCE_DOES_NOT_EXIST):
        return None
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_delete_response(response):
    if response.return_code == str(SUCCESS):
        return Status.OK
    elif response.return_code == str(INTERNAL_ERROR):
        return Status.ERROR
    else:
        raise AIFlowException(response.return_msg)


def _unwrap_update_response(response):
    if response.return_code == str(SUCCESS):
        return int(response.data)
    elif response.return_code == str(INTERNAL_ERROR):
        return Status.ERROR
    else:
        raise AIFlowException(response.return_msg)


def transform_dataset_type_list_to_proto(type_list):
    data_type_list = []
    if type_list is not None:
        for data_type in type_list:
            data_type_list.append(DataTypeProto.Value(data_type))
    else:
        data_type_list = None
    return data_type_list


def _wrap_meta_response(data):
    if data is not None:
        return Response(return_code=str(SUCCESS), return_msg=ReturnCode.Name(SUCCESS).lower(),
                        data=MessageToJson(data, preserving_proto_field_name=True))
    else:
        return Response(return_code=str(RESOURCE_DOES_NOT_EXIST),
                        return_msg=ReturnCode.Name(RESOURCE_DOES_NOT_EXIST).lower(),
                        data=None)


def _wrap_delete_response(status):
    if status == Status.OK:
        return Response(return_code=str(SUCCESS), return_msg=ReturnCode.Name(SUCCESS).lower(), data=None)
    else:
        return Response(return_code=str(INTERNAL_ERROR), return_msg=ReturnCode.Name(INTERNAL_ERROR).lower(),
                        data=None)


def _wrap_update_response(uuid):
    if uuid != UPDATE_FAIL:
        return Response(return_code=str(SUCCESS), return_msg=ReturnCode.Name(SUCCESS).lower(), data=str(uuid))
    else:
        return Response(return_code=str(INTERNAL_ERROR), return_msg=ReturnCode.Name(INTERNAL_ERROR).lower(),
                        data=None)


def _warp_dataset_list_response(dataset_meta_list):
    if dataset_meta_list is not None:
        dataset_proto_list = MetaToProto.dataset_meta_list_to_proto(dataset_meta_list)
        return Response(return_code=str(SUCCESS), return_msg=ReturnCode.Name(SUCCESS).lower(),
                        data=MessageToJson(DatasetListProto(datasets=dataset_proto_list),
                                           preserving_proto_field_name=True))
    else:
        return Response(return_code=str(RESOURCE_DOES_NOT_EXIST),
                        return_msg=ReturnCode.Name(RESOURCE_DOES_NOT_EXIST).lower(),
                        data=None)


def _warp_model_relation_list_response(model_relation_list):
    if model_relation_list is not None:
        model_proto_list = MetaToProto.model_relation_meta_list_to_proto(model_relation_list)
        return Response(return_code=str(SUCCESS), return_msg=ReturnCode.Name(SUCCESS).lower(),
                        data=MessageToJson(ModelRelationListProto(model_relations=model_proto_list),
                                           preserving_proto_field_name=True))
    else:
        return Response(return_code=str(RESOURCE_DOES_NOT_EXIST),
                        return_msg=ReturnCode.Name(RESOURCE_DOES_NOT_EXIST).lower(),
                        data=None)


def _warp_model_version_relation_list_response(model_version_relation_list):
    if model_version_relation_list is not None:
        model_version_proto_list = MetaToProto.model_version_relation_meta_list_to_proto(model_version_relation_list)
        return Response(return_code=str(SUCCESS), return_msg=ReturnCode.Name(SUCCESS).lower(),
                        data=MessageToJson(ModelVersionRelationListProto(model_versions=model_version_proto_list),
                                           preserving_proto_field_name=True))
    else:
        return Response(return_code=str(RESOURCE_DOES_NOT_EXIST),
                        return_msg=ReturnCode.Name(RESOURCE_DOES_NOT_EXIST).lower(),
                        data=None)


def _warp_project_list_response(project_list):
    if project_list is not None:
        project_proto_list = MetaToProto.project_meta_list_to_proto(project_list)
        return Response(return_code=str(SUCCESS), return_msg=ReturnCode.Name(SUCCESS).lower(),
                        data=MessageToJson(ProjectListProto(projects=project_proto_list),
                                           preserving_proto_field_name=True))
    else:
        return Response(return_code=str(RESOURCE_DOES_NOT_EXIST),
                        return_msg=ReturnCode.Name(RESOURCE_DOES_NOT_EXIST).lower(),
                        data=None)


def _wrap_workflow_list_response(workflow_list):
    if workflow_list is not None:
        workflow_proto_list = MetaToProto.workflow_meta_list_to_proto(workflow_list)
        return Response(return_code=str(SUCCESS), return_msg=ReturnCode.Name(SUCCESS).lower(),
                        data=MessageToJson(WorkflowListProto(workflows=workflow_proto_list),
                                           preserving_proto_field_name=True))
    else:
        return Response(return_code=str(RESOURCE_DOES_NOT_EXIST),
                        return_msg=ReturnCode.Name(RESOURCE_DOES_NOT_EXIST).lower(),
                        data=None)


def _warp_artifact_list_response(artifact_list):
    if artifact_list is not None:
        artifact_proto_list = MetaToProto.artifact_meta_list_to_proto(artifact_list)
        return Response(return_code=str(SUCCESS), return_msg=ReturnCode.Name(SUCCESS).lower(),
                        data=MessageToJson(ArtifactListProto(artifacts=artifact_proto_list),
                                           preserving_proto_field_name=True))
    else:
        return Response(return_code=str(RESOURCE_DOES_NOT_EXIST),
                        return_msg=ReturnCode.Name(RESOURCE_DOES_NOT_EXIST).lower(),
                        data=None)


def transform_dataset_meta(dataset_proto):
    properties = dataset_proto.properties
    if properties == {}:
        properties = None
    name_list = dataset_proto.schema.name_list
    type_list = dataset_proto.schema.type_list
    if not name_list:
        name_list = None
    if not type_list:
        data_type_list = None
    else:
        data_type_list = []
        for c in type_list:
            data_type_list.append(DataType(DataTypeProto.Name(c)))
    schema = Schema(name_list=name_list, type_list=data_type_list)
    return DatasetMeta(name=dataset_proto.name,
                       data_format=dataset_proto.data_format.value if dataset_proto.HasField('data_format') else None,
                       description=dataset_proto.description.value if dataset_proto.HasField('description') else None,
                       uri=dataset_proto.uri.value if dataset_proto.HasField('uri') else None,
                       create_time=dataset_proto.create_time.value if dataset_proto.HasField('create_time') else None,
                       update_time=dataset_proto.update_time.value if dataset_proto.HasField('update_time') else None,
                       properties=properties,
                       schema=schema,
                       catalog_name=dataset_proto.catalog_name.value if dataset_proto.HasField(
                           'catalog_name') else None,
                       catalog_type=dataset_proto.catalog_type.value if dataset_proto.HasField(
                           'catalog_type') else None,
                       catalog_database=dataset_proto.catalog_database.value if dataset_proto.HasField(
                           'catalog_database') else None,
                       catalog_connection_uri=dataset_proto.catalog_connection_uri.value \
                           if dataset_proto.HasField('catalog_connection_uri') else None,
                       catalog_table=dataset_proto.catalog_table.value if dataset_proto.HasField(
                           'catalog_table') else None)


def transform_project_meta(project_proto):
    properties = project_proto.properties
    if properties == {}:
        properties = None
    return ProjectMeta(
        name=project_proto.name,
        properties=properties,
        uri=project_proto.uri.value if project_proto.HasField('uri') else None)


def transform_workflow_meta(workflow_proto) -> WorkflowMeta:
    properties = workflow_proto.properties
    if properties == {}:
        properties = None
    return WorkflowMeta(name=workflow_proto.name,
                        project_id=workflow_proto.project_id.value if workflow_proto.HasField('project_id') else None,
                        properties=properties,
                        create_time=workflow_proto.create_time.value if workflow_proto.HasField('create_time') else None,
                        update_time=workflow_proto.update_time.value if workflow_proto.HasField('update_time') else None
                        )


def transform_artifact_meta(artifact_proto) -> ArtifactMeta:
    properties = artifact_proto.properties
    if properties == {}:
        properties = None
    return ArtifactMeta(
        name=artifact_proto.name,
        properties=properties,
        artifact_type=artifact_proto.artifact_type.value if artifact_proto.HasField('artifact_type') else None,
        description=artifact_proto.description.value if artifact_proto.HasField('description') else None,
        uri=artifact_proto.uri.value if artifact_proto.HasField('uri') else None,
        create_time=artifact_proto.create_time.value if artifact_proto.HasField('create_time') else None,
        update_time=artifact_proto.update_time.value if artifact_proto.HasField('update_time') else None)


def transform_model_relation_meta(model_relation_proto):
    return ModelRelationMeta(name=model_relation_proto.name,
                             project_id=model_relation_proto.project_id.value if model_relation_proto.HasField(
                                 'project_id') else None)


def transform_model_version_relation_meta(model_version_relation_proto):
    return ModelVersionRelationMeta(
        version=model_version_relation_proto.version.value
        if model_version_relation_proto.HasField('version') else None,
        model_id=model_version_relation_proto.model_id.value
        if model_version_relation_proto.HasField('model_id') else None,
        project_snapshot_id=model_version_relation_proto.project_snapshot_id.value
        if model_version_relation_proto.HasField('project_snapshot_id') else None)


def transform_model_meta(model_proto):
    return ModelMeta(name=model_proto.name,
                     model_desc=model_proto.model_desc.value if model_proto.HasField('model_desc') else None,
                     project_id=model_proto.project_id.value if model_proto.HasField('project_id') else None)


def transform_model_version_meta(model_version_proto):
    return ModelVersionMeta(version=model_version_proto.version.value if model_version_proto.HasField(
        'version') else None,
                            model_id=model_version_proto.model_id.value if model_version_proto.HasField(
                                'model_id') else None,
                            project_snapshot_id=model_version_proto.project_snapshot_id.value
                            if model_version_proto.HasField('project_snapshot_id') else None,
                            model_path=model_version_proto.model_path.value if model_version_proto.HasField(
                                "model_path") else None,
                            model_type=model_version_proto.model_type.value if model_version_proto.HasField(
                                "model_type") else None,
                            version_desc=model_version_proto.version_desc.value if model_version_proto.HasField(
                                "version_desc") else None)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module is deprecated. Please use `airflow.providers.google.cloud.operators.vision`."""

import warnings

from airflow.providers.google.cloud.operators.vision import (  # noqa # pylint: disable=unused-import
    CloudVisionAddProductToProductSetOperator,
    CloudVisionCreateProductOperator,
    CloudVisionCreateProductSetOperator,
    CloudVisionCreateReferenceImageOperator,
    CloudVisionDeleteProductOperator,
    CloudVisionDeleteProductSetOperator,
    CloudVisionDetectImageLabelsOperator,
    CloudVisionDetectImageSafeSearchOperator,
    CloudVisionDetectTextOperator,
    CloudVisionGetProductOperator,
    CloudVisionGetProductSetOperator,
    CloudVisionImageAnnotateOperator,
    CloudVisionRemoveProductFromProductSetOperator,
    CloudVisionTextDetectOperator,
    CloudVisionUpdateProductOperator,
    CloudVisionUpdateProductSetOperator,
)

warnings.warn(
    "This module is deprecated. Please use `airflow.providers.google.cloud.operators.vision`.",
    DeprecationWarning,
    stacklevel=2,
)


class CloudVisionAnnotateImageOperator(CloudVisionImageAnnotateOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.vision.CloudVisionImageAnnotateOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.vision.CloudVisionImageAnnotateOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class CloudVisionDetectDocumentTextOperator(CloudVisionTextDetectOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.vision.CloudVisionTextDetectOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.vision.CloudVisionTextDetectOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class CloudVisionProductCreateOperator(CloudVisionCreateProductOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.vision.CloudVisionCreateProductOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.vision.CloudVisionCreateProductOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class CloudVisionProductDeleteOperator(CloudVisionDeleteProductOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.vision.CloudVisionDeleteProductOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.vision.CloudVisionDeleteProductOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class CloudVisionProductGetOperator(CloudVisionGetProductOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.vision.CloudVisionGetProductOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use `airflow.providers.google.cloud.operators.vision.CloudVisionGetProductOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class CloudVisionProductSetCreateOperator(CloudVisionCreateProductSetOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.vision.CloudVisionCreateProductSetOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use
            `airflow.providers.google.cloud.operators.vision.CloudVisionCreateProductSetOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class CloudVisionProductSetDeleteOperator(CloudVisionDeleteProductSetOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.vision.CloudVisionDeleteProductSetOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use
            `airflow.providers.google.cloud.operators.vision.CloudVisionDeleteProductSetOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class CloudVisionProductSetGetOperator(CloudVisionGetProductSetOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.vision.CloudVisionGetProductSetOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use
            `airflow.providers.google.cloud.operators.vision.CloudVisionGetProductSetOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class CloudVisionProductSetUpdateOperator(CloudVisionUpdateProductSetOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.vision.CloudVisionUpdateProductSetOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use
            `airflow.providers.google.cloud.operators.vision.CloudVisionUpdateProductSetOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class CloudVisionProductUpdateOperator(CloudVisionUpdateProductOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.vision.CloudVisionUpdateProductOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use
            `airflow.providers.google.cloud.operators.vision.CloudVisionUpdateProductOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)


class CloudVisionReferenceImageCreateOperator(CloudVisionCreateReferenceImageOperator):
    """
    This class is deprecated.
    Please use `airflow.providers.google.cloud.operators.vision.CloudVisionCreateReferenceImageOperator`.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            """This class is deprecated.
            Please use
            `airflow.providers.google.cloud.operators.vision.CloudVisionCreateReferenceImageOperator`.""",
            DeprecationWarning,
            stacklevel=3,
        )
        super().__init__(*args, **kwargs)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
from typing import NamedTuple, Optional

import airflow
from airflow.api_connexion.schemas.version_schema import version_info_schema
from airflow.utils.platform import get_airflow_git_version

log = logging.getLogger(__name__)


class VersionInfo(NamedTuple):
    """Version information"""

    version: str
    git_version: Optional[str]


def get_version():
    """Get version information"""
    airflow_version = airflow.__version__

    git_version = get_airflow_git_version()
    version_info = VersionInfo(version=airflow_version, git_version=git_version)
    return version_info_schema.dump(version_info)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from ai_flow.meta.job_meta import ExecutionMode
from ai_flow.util.json_utils import Jsonable
from typing import Text, Dict, Optional


class JobConfig(Jsonable):
    """
    JobConfig is the configuration information of the Job(ai_flow.workflow.job.Job).
    """

    def __init__(self,
                 job_name: Text = None,
                 job_type: Text = None,
                 properties: Dict[Text, Jsonable] = None) -> None:
        """
        :param job_name: Name of the configured job.
        :param job_type: The job type of the job, such as bash python etc..
        :param properties: Properties of the configured job.
        """
        super().__init__()
        self.job_name: Text = job_name
        self.job_type = job_type
        if properties is None:
            self.properties: Dict[Text, Jsonable] = {}
        else:
            self.properties: Dict[Text, Jsonable] = properties

    @staticmethod
    def from_dict(data: Dict) -> 'JobConfig':
        """
        Build JobConfig from a dict.
        The dict example:
        {'job_name':{'job_type':'bash', 'properties': {}}}
        """
        job_name = list(data.keys())[0]
        return JobConfig(job_name=job_name,
                         job_type=data[job_name].get('job_type', None),
                         properties=data[job_name].get('properties', {}))

    @staticmethod
    def to_dict(job_config: 'JobConfig') -> Dict:
        """Convert the JobConfig to a dict"""
        return {job_config.job_name: {'job_type': job_config.job_type, 'properties': job_config.properties}}



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from airflow.ti_deps.deps.base_ti_dep import BaseTIDep
from airflow.utils import timezone
from airflow.utils.session import provide_session


class RunnableExecDateDep(BaseTIDep):
    """Determines whether a task's execution date is valid."""

    NAME = "Execution Date"
    IGNOREABLE = True

    @provide_session
    def _get_dep_statuses(self, ti, session, dep_context):
        cur_date = timezone.utcnow()

        # don't consider runs that are executed in the future unless
        # specified by config and schedule_interval is None
        if ti.execution_date > cur_date and not ti.task.dag.allow_future_exec_dates:
            yield self._failing_status(
                reason="Execution date {} is in the future (the current "
                "date is {}).".format(ti.execution_date.isoformat(), cur_date.isoformat())
            )

        if ti.task.end_date and ti.execution_date > ti.task.end_date:
            yield self._failing_status(
                reason="The execution date is {} but this is after the task's end date "
                "{}.".format(ti.execution_date.isoformat(), ti.task.end_date.isoformat())
            )

        if ti.task.dag and ti.task.dag.end_date and ti.execution_date > ti.task.dag.end_date:
            yield self._failing_status(
                reason="The execution date is {} but this is after the task's DAG's "
                "end date {}.".format(ti.execution_date.isoformat(), ti.task.dag.end_date.isoformat())
            )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Manages all plugins."""
import importlib
import importlib.machinery
import importlib.util
import inspect
import logging
import os
import sys
import types
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type

try:
    import importlib_metadata
except ImportError:
    from importlib import metadata as importlib_metadata

from airflow import settings
from airflow.utils.entry_points import entry_points_with_dist
from airflow.utils.file import find_path_from_directory

if TYPE_CHECKING:
    from airflow.hooks.base import BaseHook

log = logging.getLogger(__name__)

import_errors: Dict[str, str] = {}

plugins = None  # type: Optional[List[AirflowPlugin]]

# Plugin components to integrate as modules
registered_hooks: Optional[List['BaseHook']] = None
macros_modules: Optional[List[Any]] = None
executors_modules: Optional[List[Any]] = None

# Plugin components to integrate directly
admin_views: Optional[List[Any]] = None
flask_blueprints: Optional[List[Any]] = None
menu_links: Optional[List[Any]] = None
flask_appbuilder_views: Optional[List[Any]] = None
flask_appbuilder_menu_links: Optional[List[Any]] = None
global_operator_extra_links: Optional[List[Any]] = None
operator_extra_links: Optional[List[Any]] = None
registered_operator_link_classes: Optional[Dict[str, Type]] = None
"""Mapping of class names to class of OperatorLinks registered by plugins.

Used by the DAG serialization code to only allow specific classes to be created
during deserialization
"""


class AirflowPluginSource:
    """Class used to define an AirflowPluginSource."""

    def __str__(self):
        raise NotImplementedError

    def __html__(self):
        raise NotImplementedError


class PluginsDirectorySource(AirflowPluginSource):
    """Class used to define Plugins loaded from Plugins Directory."""

    def __init__(self, path):
        self.path = os.path.relpath(path, settings.PLUGINS_FOLDER)

    def __str__(self):
        return f"$PLUGINS_FOLDER/{self.path}"

    def __html__(self):
        return f"<em>$PLUGINS_FOLDER/</em>{self.path}"


class EntryPointSource(AirflowPluginSource):
    """Class used to define Plugins loaded from entrypoint."""

    def __init__(self, entrypoint: importlib_metadata.EntryPoint, dist: importlib_metadata.Distribution):
        self.dist = dist.metadata['name']
        self.version = dist.version
        self.entrypoint = str(entrypoint)

    def __str__(self):
        return f"{self.dist}=={self.version}: {self.entrypoint}"

    def __html__(self):
        return f"<em>{self.dist}=={self.version}:</em> {self.entrypoint}"


class AirflowPluginException(Exception):
    """Exception when loading plugin."""


class AirflowPlugin:
    """Class used to define AirflowPlugin."""

    name: Optional[str] = None
    source: Optional[AirflowPluginSource] = None
    hooks: List[Any] = []
    executors: List[Any] = []
    macros: List[Any] = []
    admin_views: List[Any] = []
    flask_blueprints: List[Any] = []
    menu_links: List[Any] = []
    appbuilder_views: List[Any] = []
    appbuilder_menu_items: List[Any] = []

    # A list of global operator extra links that can redirect users to
    # external systems. These extra links will be available on the
    # task page in the form of buttons.
    #
    # Note: the global operator extra link can be overridden at each
    # operator level.
    global_operator_extra_links: List[Any] = []

    # A list of operator extra links to override or add operator links
    # to existing Airflow Operators.
    # These extra links will be available on the task page in form of
    # buttons.
    operator_extra_links: List[Any] = []

    @classmethod
    def validate(cls):
        """Validates that plugin has a name."""
        if not cls.name:
            raise AirflowPluginException("Your plugin needs a name.")

    @classmethod
    def on_load(cls, *args, **kwargs):
        """
        Executed when the plugin is loaded.
        This method is only called once during runtime.

        :param args: If future arguments are passed in on call.
        :param kwargs: If future arguments are passed in on call.
        """


def is_valid_plugin(plugin_obj):
    """
    Check whether a potential object is a subclass of
    the AirflowPlugin class.

    :param plugin_obj: potential subclass of AirflowPlugin
    :return: Whether or not the obj is a valid subclass of
        AirflowPlugin
    """
    global plugins  # pylint: disable=global-statement

    if (
        inspect.isclass(plugin_obj)
        and issubclass(plugin_obj, AirflowPlugin)
        and (plugin_obj is not AirflowPlugin)
    ):
        plugin_obj.validate()
        return plugin_obj not in plugins
    return False


def load_entrypoint_plugins():
    """
    Load and register plugins AirflowPlugin subclasses from the entrypoints.
    The entry_point group should be 'airflow.plugins'.
    """
    global import_errors  # pylint: disable=global-statement
    global plugins  # pylint: disable=global-statement

    log.debug("Loading plugins from entrypoints")

    for entry_point, dist in entry_points_with_dist('airflow.plugins'):
        log.debug('Importing entry_point plugin %s', entry_point.name)
        try:
            plugin_class = entry_point.load()
            if not is_valid_plugin(plugin_class):
                continue

            plugin_instance = plugin_class()
            if callable(getattr(plugin_instance, 'on_load', None)):
                plugin_instance.on_load()
                plugin_instance.source = EntryPointSource(entry_point, dist)
                plugins.append(plugin_instance)
        except Exception as e:  # pylint: disable=broad-except
            log.exception("Failed to import plugin %s", entry_point.name)
            import_errors[entry_point.module] = str(e)


def load_plugins_from_plugin_directory():
    """Load and register Airflow Plugins from plugins directory"""
    global import_errors  # pylint: disable=global-statement
    global plugins  # pylint: disable=global-statement
    log.debug("Loading plugins from directory: %s", settings.PLUGINS_FOLDER)

    for file_path in find_path_from_directory(settings.PLUGINS_FOLDER, ".airflowignore"):

        if not os.path.isfile(file_path):
            continue
        mod_name, file_ext = os.path.splitext(os.path.split(file_path)[-1])
        if file_ext != '.py':
            continue

        try:
            loader = importlib.machinery.SourceFileLoader(mod_name, file_path)
            spec = importlib.util.spec_from_loader(mod_name, loader)
            mod = importlib.util.module_from_spec(spec)
            sys.modules[spec.name] = mod
            loader.exec_module(mod)
            log.debug('Importing plugin module %s', file_path)

            for mod_attr_value in (m for m in mod.__dict__.values() if is_valid_plugin(m)):
                plugin_instance = mod_attr_value()
                plugin_instance.source = PluginsDirectorySource(file_path)
                plugins.append(plugin_instance)

        except Exception as e:  # pylint: disable=broad-except
            log.exception(e)
            log.error('Failed to import plugin %s', file_path)
            import_errors[file_path] = str(e)


# pylint: disable=protected-access
def make_module(name: str, objects: List[Any]):
    """Creates new module."""
    if not objects:
        return None
    log.debug('Creating module %s', name)
    name = name.lower()
    module = types.ModuleType(name)
    module._name = name.split('.')[-1]  # type: ignore
    module._objects = objects  # type: ignore
    module.__dict__.update((o.__name__, o) for o in objects)
    return module


# pylint: enable=protected-access


def ensure_plugins_loaded():
    """
    Load plugins from plugins directory and entrypoints.

    Plugins are only loaded if they have not been previously loaded.
    """
    from airflow.stats import Stats

    global plugins, registered_hooks  # pylint: disable=global-statement

    if plugins is not None:
        log.debug("Plugins are already loaded. Skipping.")
        return

    if not settings.PLUGINS_FOLDER:
        raise ValueError("Plugins folder is not set")

    log.debug("Loading plugins")

    with Stats.timer() as timer:
        plugins = []
        registered_hooks = []

        load_plugins_from_plugin_directory()
        load_entrypoint_plugins()

        # We don't do anything with these for now, but we want to keep track of
        # them so we can integrate them in to the UI's Connection screens
        for plugin in plugins:
            registered_hooks.extend(plugin.hooks)

    num_loaded = len(plugins)
    if num_loaded > 0:
        log.info("Loading %d plugin(s) took %.2f seconds", num_loaded, timer.duration)


def initialize_web_ui_plugins():
    """Collect extension points for WEB UI"""
    # pylint: disable=global-statement
    global plugins
    global flask_blueprints
    global flask_appbuilder_views
    global flask_appbuilder_menu_links
    # pylint: enable=global-statement

    if (
        flask_blueprints is not None
        and flask_appbuilder_views is not None
        and flask_appbuilder_menu_links is not None
    ):
        return

    ensure_plugins_loaded()

    if plugins is None:
        raise AirflowPluginException("Can't load plugins.")

    log.debug("Initialize Web UI plugin")

    flask_blueprints = []
    flask_appbuilder_views = []
    flask_appbuilder_menu_links = []

    for plugin in plugins:
        flask_appbuilder_views.extend(plugin.appbuilder_views)
        flask_appbuilder_menu_links.extend(plugin.appbuilder_menu_items)
        flask_blueprints.extend([{'name': plugin.name, 'blueprint': bp} for bp in plugin.flask_blueprints])

        if (plugin.admin_views and not plugin.appbuilder_views) or (
            plugin.menu_links and not plugin.appbuilder_menu_items
        ):
            log.warning(
                "Plugin \'%s\' may not be compatible with the current Airflow version. "
                "Please contact the author of the plugin.",
                plugin.name,
            )


def initialize_extra_operators_links_plugins():
    """Creates modules for loaded extension from extra operators links plugins"""
    # pylint: disable=global-statement
    global global_operator_extra_links
    global operator_extra_links
    global registered_operator_link_classes
    # pylint: enable=global-statement

    if (
        global_operator_extra_links is not None
        and operator_extra_links is not None
        and registered_operator_link_classes is not None
    ):
        return

    ensure_plugins_loaded()

    if plugins is None:
        raise AirflowPluginException("Can't load plugins.")

    log.debug("Initialize extra operators links plugins")

    global_operator_extra_links = []
    operator_extra_links = []
    registered_operator_link_classes = {}

    for plugin in plugins:
        global_operator_extra_links.extend(plugin.global_operator_extra_links)
        operator_extra_links.extend(list(plugin.operator_extra_links))

        registered_operator_link_classes.update(
            {
                f"{link.__class__.__module__}.{link.__class__.__name__}": link.__class__
                for link in plugin.operator_extra_links
            }
        )


def integrate_executor_plugins() -> None:
    """Integrate executor plugins to the context."""
    # pylint: disable=global-statement
    global plugins
    global executors_modules
    # pylint: enable=global-statement

    if executors_modules is not None:
        return

    ensure_plugins_loaded()

    if plugins is None:
        raise AirflowPluginException("Can't load plugins.")

    log.debug("Integrate executor plugins")

    executors_modules = []
    for plugin in plugins:
        if plugin.name is None:
            raise AirflowPluginException("Invalid plugin name")
        plugin_name: str = plugin.name

        executors_module = make_module('airflow.executors.' + plugin_name, plugin.executors)
        if executors_module:
            executors_modules.append(executors_module)
            sys.modules[executors_module.__name__] = executors_module  # pylint: disable=no-member


def integrate_macros_plugins() -> None:
    """Integrates macro plugins."""
    # pylint: disable=global-statement
    global plugins
    global macros_modules
    # pylint: enable=global-statement
    from airflow import macros

    if macros_modules is not None:
        return

    ensure_plugins_loaded()

    if plugins is None:
        raise AirflowPluginException("Can't load plugins.")

    log.debug("Integrate DAG plugins")

    macros_modules = []

    for plugin in plugins:
        if plugin.name is None:
            raise AirflowPluginException("Invalid plugin name")

        macros_module = make_module(f'airflow.macros.{plugin.name}', plugin.macros)

        if macros_module:
            macros_modules.append(macros_module)
            sys.modules[macros_module.__name__] = macros_module  # pylint: disable=no-member
            # Register the newly created module on airflow.macros such that it
            # can be accessed when rendering templates.
            setattr(macros, plugin.name, macros_module)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG that shows the complex DAG structure.
"""

from airflow import models
from airflow.models.baseoperator import chain
from airflow.operators.bash import BashOperator
from airflow.operators.python import PythonOperator
from airflow.utils.dates import days_ago

with models.DAG(
    dag_id="example_complex",
    schedule_interval=None,
    start_date=days_ago(1),
    tags=['example', 'example2', 'example3'],
) as dag:

    # Create
    create_entry_group = BashOperator(task_id="create_entry_group", bash_command="echo create_entry_group")

    create_entry_group_result = BashOperator(
        task_id="create_entry_group_result", bash_command="echo create_entry_group_result"
    )

    create_entry_group_result2 = BashOperator(
        task_id="create_entry_group_result2", bash_command="echo create_entry_group_result2"
    )

    create_entry_gcs = BashOperator(task_id="create_entry_gcs", bash_command="echo create_entry_gcs")

    create_entry_gcs_result = BashOperator(
        task_id="create_entry_gcs_result", bash_command="echo create_entry_gcs_result"
    )

    create_entry_gcs_result2 = BashOperator(
        task_id="create_entry_gcs_result2", bash_command="echo create_entry_gcs_result2"
    )

    create_tag = BashOperator(task_id="create_tag", bash_command="echo create_tag")

    create_tag_result = BashOperator(task_id="create_tag_result", bash_command="echo create_tag_result")

    create_tag_result2 = BashOperator(task_id="create_tag_result2", bash_command="echo create_tag_result2")

    create_tag_template = BashOperator(task_id="create_tag_template", bash_command="echo create_tag_template")

    create_tag_template_result = BashOperator(
        task_id="create_tag_template_result", bash_command="echo create_tag_template_result"
    )

    create_tag_template_result2 = BashOperator(
        task_id="create_tag_template_result2", bash_command="echo create_tag_template_result2"
    )

    create_tag_template_field = BashOperator(
        task_id="create_tag_template_field", bash_command="echo create_tag_template_field"
    )

    create_tag_template_field_result = BashOperator(
        task_id="create_tag_template_field_result", bash_command="echo create_tag_template_field_result"
    )

    create_tag_template_field_result2 = BashOperator(
        task_id="create_tag_template_field_result2", bash_command="echo create_tag_template_field_result"
    )

    # Delete
    delete_entry = BashOperator(task_id="delete_entry", bash_command="echo delete_entry")
    create_entry_gcs >> delete_entry

    delete_entry_group = BashOperator(task_id="delete_entry_group", bash_command="echo delete_entry_group")
    create_entry_group >> delete_entry_group

    delete_tag = BashOperator(task_id="delete_tag", bash_command="echo delete_tag")
    create_tag >> delete_tag

    delete_tag_template_field = BashOperator(
        task_id="delete_tag_template_field", bash_command="echo delete_tag_template_field"
    )

    delete_tag_template = BashOperator(task_id="delete_tag_template", bash_command="echo delete_tag_template")

    # Get
    get_entry_group = BashOperator(task_id="get_entry_group", bash_command="echo get_entry_group")

    get_entry_group_result = BashOperator(
        task_id="get_entry_group_result", bash_command="echo get_entry_group_result"
    )

    get_entry = BashOperator(task_id="get_entry", bash_command="echo get_entry")

    get_entry_result = BashOperator(task_id="get_entry_result", bash_command="echo get_entry_result")

    get_tag_template = BashOperator(task_id="get_tag_template", bash_command="echo get_tag_template")

    get_tag_template_result = BashOperator(
        task_id="get_tag_template_result", bash_command="echo get_tag_template_result"
    )

    # List
    list_tags = BashOperator(task_id="list_tags", bash_command="echo list_tags")

    list_tags_result = BashOperator(task_id="list_tags_result", bash_command="echo list_tags_result")

    # Lookup
    lookup_entry = BashOperator(task_id="lookup_entry", bash_command="echo lookup_entry")

    lookup_entry_result = BashOperator(task_id="lookup_entry_result", bash_command="echo lookup_entry_result")

    # Rename
    rename_tag_template_field = BashOperator(
        task_id="rename_tag_template_field", bash_command="echo rename_tag_template_field"
    )

    # Search
    search_catalog = PythonOperator(task_id="search_catalog", python_callable=lambda: print("search_catalog"))

    search_catalog_result = BashOperator(
        task_id="search_catalog_result", bash_command="echo search_catalog_result"
    )

    # Update
    update_entry = BashOperator(task_id="update_entry", bash_command="echo update_entry")

    update_tag = BashOperator(task_id="update_tag", bash_command="echo update_tag")

    update_tag_template = BashOperator(task_id="update_tag_template", bash_command="echo update_tag_template")

    update_tag_template_field = BashOperator(
        task_id="update_tag_template_field", bash_command="echo update_tag_template_field"
    )

    # Create
    create_tasks = [
        create_entry_group,
        create_entry_gcs,
        create_tag_template,
        create_tag_template_field,
        create_tag,
    ]
    chain(*create_tasks)

    create_entry_group >> delete_entry_group
    create_entry_group >> create_entry_group_result
    create_entry_group >> create_entry_group_result2

    create_entry_gcs >> delete_entry
    create_entry_gcs >> create_entry_gcs_result
    create_entry_gcs >> create_entry_gcs_result2

    create_tag_template >> delete_tag_template_field
    create_tag_template >> create_tag_template_result
    create_tag_template >> create_tag_template_result2

    create_tag_template_field >> delete_tag_template_field
    create_tag_template_field >> create_tag_template_field_result
    create_tag_template_field >> create_tag_template_field_result2

    create_tag >> delete_tag
    create_tag >> create_tag_result
    create_tag >> create_tag_result2

    # Delete
    delete_tasks = [
        delete_tag,
        delete_tag_template_field,
        delete_tag_template,
        delete_entry_group,
        delete_entry,
    ]
    chain(*delete_tasks)

    # Get
    create_tag_template >> get_tag_template >> delete_tag_template
    get_tag_template >> get_tag_template_result

    create_entry_gcs >> get_entry >> delete_entry
    get_entry >> get_entry_result

    create_entry_group >> get_entry_group >> delete_entry_group
    get_entry_group >> get_entry_group_result

    # List
    create_tag >> list_tags >> delete_tag
    list_tags >> list_tags_result

    # Lookup
    create_entry_gcs >> lookup_entry >> delete_entry
    lookup_entry >> lookup_entry_result

    # Rename
    create_tag_template_field >> rename_tag_template_field >> delete_tag_template_field

    # Search
    chain(create_tasks, search_catalog, delete_tasks)
    search_catalog >> search_catalog_result

    # Update
    create_entry_gcs >> update_entry >> delete_entry
    create_tag >> update_tag >> delete_tag
    create_tag_template >> update_tag_template >> delete_tag_template
    create_tag_template_field >> update_tag_template_field >> rename_tag_template_field



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import warnings
from abc import ABC
from typing import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    from airflow.models.connection import Connection


class BaseSecretsBackend(ABC):
    """Abstract base class to retrieve Connection object given a conn_id or Variable given a key"""

    def __init__(self, **kwargs):
        pass

    @staticmethod
    def build_path(path_prefix: str, secret_id: str, sep: str = "/") -> str:
        """
        Given conn_id, build path for Secrets Backend

        :param path_prefix: Prefix of the path to get secret
        :type path_prefix: str
        :param secret_id: Secret id
        :type secret_id: str
        :param sep: separator used to concatenate connections_prefix and conn_id. Default: "/"
        :type sep: str
        """
        return f"{path_prefix}{sep}{secret_id}"

    def get_conn_uri(self, conn_id: str) -> Optional[str]:
        """
        Get conn_uri from Secrets Backend

        :param conn_id: connection id
        :type conn_id: str
        """
        raise NotImplementedError()

    def get_connection(self, conn_id: str) -> Optional['Connection']:
        """
        Return connection object with a given ``conn_id``.

        :param conn_id: connection id
        :type conn_id: str
        """
        from airflow.models.connection import Connection

        conn_uri = self.get_conn_uri(conn_id=conn_id)
        if not conn_uri:
            return None
        conn = Connection(conn_id=conn_id, uri=conn_uri)
        return conn

    def get_connections(self, conn_id: str) -> List['Connection']:
        """
        Return connection object with a given ``conn_id``.

        :param conn_id: connection id
        :type conn_id: str
        """
        warnings.warn(
            "This method is deprecated. Please use "
            "`airflow.secrets.base_secrets.BaseSecretsBackend.get_connection`.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        conn = self.get_connection(conn_id=conn_id)
        if conn:
            return [conn]
        return []

    def get_variable(self, key: str) -> Optional[str]:
        """
        Return value for Airflow Variable

        :param key: Variable Key
        :type key: str
        :return: Variable Value
        """
        raise NotImplementedError()

    def get_config(self, key: str) -> Optional[str]:  # pylint: disable=unused-argument
        """
        Return value for Airflow Config Key

        :param key: Config Key
        :return: Config Value
        """
        return None



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Text, List, Dict

import grpc
from ai_flow.protobuf import scheduling_service_pb2_grpc
from ai_flow.protobuf import scheduling_service_pb2
from ai_flow.protobuf.message_pb2 import WorkflowProto, WorkflowExecutionProto, JobProto, StatusProto

from ai_flow.endpoint.client.base_client import BaseClient


class SchedulerClient(BaseClient):
    def __init__(self, server_uri):
        super(SchedulerClient, self).__init__(server_uri)
        channel = grpc.insecure_channel(server_uri)
        self.scheduling_stub = scheduling_service_pb2_grpc.SchedulingServiceStub(channel)

    def submit_workflow_to_scheduler(self,
                                     namespace: Text,
                                     workflow_json: Text,
                                     workflow_name: Text = None,
                                     args: Dict = None) -> WorkflowProto:
        """
        Submit the ai flow workflow to the scheduler.
        :param workflow_json:
        :param namespace:
        :param workflow_name: The ai flow workflow identify.
        :param args: The arguments of the submit action.
        :return: The result of the submit action.
        """
        request = scheduling_service_pb2.ScheduleWorkflowRequest()
        request.namespace = namespace
        request.workflow_name = workflow_name
        request.workflow_json = workflow_json
        if args is not None:
            for k, v in args.items():
                request.args[k] = v
        response = self.scheduling_stub.submitWorkflow(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.workflow

    def delete_workflow(self,
                        namespace: Text,
                        workflow_name: Text = None) -> WorkflowProto:
        """
        Delete the ai flow workflow from the scheduler.
        :param namespace:
        :param workflow_name: The ai flow workflow identify.
        :return: The result of the action.
        """
        request = scheduling_service_pb2.ScheduleWorkflowRequest()
        request.namespace = namespace
        request.workflow_name = workflow_name
        response = self.scheduling_stub.deleteWorkflow(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.workflow

    def pause_workflow_scheduling(self,
                                  namespace: Text,
                                  workflow_name: Text = None) -> WorkflowProto:
        """
        Pause the ai flow workflow from the scheduler.
        :param namespace:
        :param workflow_name: The ai flow workflow identify.
        :return: The result of the action.
        """
        request = scheduling_service_pb2.ScheduleWorkflowRequest()
        request.namespace = namespace
        request.workflow_name = workflow_name
        response = self.scheduling_stub.pauseWorkflowScheduling(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.workflow

    def resume_workflow_scheduling(self,
                                   namespace: Text,
                                   workflow_name: Text = None) -> WorkflowProto:
        """
        Resume the ai flow workflow from the scheduler.
        :param namespace:
        :param workflow_name: The ai flow workflow identify.
        :return: The result of the action.
        """
        request = scheduling_service_pb2.ScheduleWorkflowRequest()
        request.namespace = namespace
        request.workflow_name = workflow_name
        response = self.scheduling_stub.resumeWorkflowScheduling(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.workflow

    def get_workflow(self,
                     namespace: Text,
                     workflow_name: Text = None) -> WorkflowProto:
        """
        Return the workflow information.
        :param namespace:
        :param workflow_name: The ai flow workflow identify.
        :return: the workflow information.
        """
        request = scheduling_service_pb2.ScheduleWorkflowRequest()
        request.namespace = namespace
        request.workflow_name = workflow_name
        response = self.scheduling_stub.getWorkflow(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.workflow

    def list_workflows(self, namespace: Text) -> List[WorkflowProto]:
        """
        :return: All workflow information.
        """
        request = scheduling_service_pb2.ScheduleWorkflowRequest()
        request.namespace = namespace
        response = self.scheduling_stub.listWorkflows(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.workflow_list

    def start_new_workflow_execution(self,
                                     namespace: Text,
                                     workflow_name: Text) -> WorkflowExecutionProto:
        """
        Run the project under the current project path.
        :param namespace:
        :param workflow_name: The ai flow workflow identify.
        :return: The result of the run action.
        """
        request = scheduling_service_pb2.WorkflowExecutionRequest()
        request.namespace = namespace
        request.workflow_name = workflow_name
        response = self.scheduling_stub.startNewWorkflowExecution(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.workflow_execution

    def kill_all_workflow_executions(self,
                                     namespace: Text,
                                     workflow_name: Text) -> List[WorkflowExecutionProto]:
        """
        Stop all instances of the workflow.
        :param namespace:
        :param workflow_name: The ai flow workflow identify.
        :return: The result of the action.
        """
        request = scheduling_service_pb2.WorkflowExecutionRequest()
        request.namespace = namespace
        request.workflow_name = workflow_name
        response = self.scheduling_stub.killAllWorkflowExecutions(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.workflow_execution_list

    def kill_workflow_execution(self,
                                execution_id: Text) -> WorkflowExecutionProto:
        """
        Stop the instance of the workflow.
        :param execution_id: The ai flow workflow execution identify.
        :return: The result of the action.
        """
        request = scheduling_service_pb2.WorkflowExecutionRequest()
        request.execution_id = execution_id
        response = self.scheduling_stub.killWorkflowExecution(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.workflow_execution

    def get_workflow_execution(self,
                               execution_id: Text) -> WorkflowExecutionProto:
        """
        Get the WorkflowExecutionInfo from scheduler.
        :param execution_id:
        :return: WorkflowExecutionInfo
        """
        request = scheduling_service_pb2.WorkflowExecutionRequest()
        request.execution_id = execution_id
        response = self.scheduling_stub.getWorkflowExecution(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.workflow_execution

    def list_workflow_executions(self,
                                 namespace: Text,
                                 workflow_name: Text) -> List[WorkflowExecutionProto]:
        """
        :param namespace:
        :param workflow_name: The ai flow workflow identify.
        :return: All workflow executions of the workflow.
        """
        request = scheduling_service_pb2.WorkflowExecutionRequest()
        request.namespace = namespace
        request.workflow_name = workflow_name
        response = self.scheduling_stub.listWorkflowExecutions(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.workflow_execution_list

    def start_job(self, job_name: Text,
                  execution_id: Text) -> JobProto:
        """
        Start a job defined in the ai flow workflow.
        :param job_name: The job name which task defined in workflow.
        :param execution_id: The ai flow workflow execution identify.
        :return: The result of the action.
        """
        request = scheduling_service_pb2.ScheduleJobRequest()
        request.job_name = job_name
        request.execution_id = execution_id
        response = self.scheduling_stub.startJob(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.job

    def stop_job(self, job_name: Text,
                 execution_id: Text) -> JobProto:
        """
        Stop a job defined in the ai flow workflow.
        :param job_name: The job name which task defined in workflow.
        :param execution_id: The ai flow workflow execution identify.
        :return: The result of the action.
        """
        request = scheduling_service_pb2.ScheduleJobRequest()
        request.job_name = job_name
        request.execution_id = execution_id
        response = self.scheduling_stub.stopJob(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.job

    def restart_job(self, job_name: Text,
                    execution_id: Text) -> JobProto:
        """
        Restart a task defined in the ai flow workflow.
        :param job_name: The job name which task defined in workflow.
        :param execution_id: The ai flow workflow execution identify.
        :return: The result of the action.
        """
        request = scheduling_service_pb2.ScheduleJobRequest()
        request.job_name = job_name
        request.execution_id = execution_id
        response = self.scheduling_stub.restartJob(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.job

    def get_job(self, job_name: Text,
                execution_id: Text) -> JobProto:
        """
        Get job information by job name.
        :param job_name:
        :param execution_id:
        :return:
        """
        request = scheduling_service_pb2.ScheduleJobRequest()
        request.job_name = job_name
        request.execution_id = execution_id
        response = self.scheduling_stub.getJob(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.job

    def list_jobs(self, execution_id: Text) -> List[JobProto]:
        """
        List the jobs of the workflow execution.
        :param execution_id:
        :return:
        """
        request = scheduling_service_pb2.ScheduleJobRequest()
        request.execution_id = execution_id
        response = self.scheduling_stub.listJobs(request)
        if response.result.status != StatusProto.OK:
            raise Exception(response.result.error_message)
        return response.job_list




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Smart sensor DAGs managing all smart sensor tasks."""
from datetime import timedelta

from airflow.configuration import conf
from airflow.models import DAG
from airflow.sensors.smart_sensor import SmartSensorOperator
from airflow.utils.dates import days_ago

args = {
    'owner': 'airflow',
}

num_smart_sensor_shard = conf.getint("smart_sensor", "shards")
shard_code_upper_limit = conf.getint('smart_sensor', 'shard_code_upper_limit')

for i in range(num_smart_sensor_shard):
    shard_min = (i * shard_code_upper_limit) / num_smart_sensor_shard
    shard_max = ((i + 1) * shard_code_upper_limit) / num_smart_sensor_shard

    dag_id = f'smart_sensor_group_shard_{i}'
    dag = DAG(
        dag_id=dag_id,
        default_args=args,
        schedule_interval=timedelta(minutes=5),
        concurrency=1,
        max_active_runs=1,
        catchup=False,
        dagrun_timeout=timedelta(hours=24),
        start_date=days_ago(2),
    )

    SmartSensorOperator(
        task_id='smart_sensor_task',
        dag=dag,
        retries=100,
        retry_delay=timedelta(seconds=10),
        priority_weight=999,
        shard_min=shard_min,
        shard_max=shard_max,
        poke_timeout=10,
        smart_sensor_timeout=timedelta(hours=24).total_seconds(),
    )

    globals()[dag_id] = dag



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import functools
import gzip
from io import BytesIO as IO
from typing import Callable, TypeVar, cast

import pendulum
from flask import after_this_request, g, request

from airflow.models import Log
from airflow.utils.session import create_session

T = TypeVar("T", bound=Callable)  # pylint: disable=invalid-name


def action_logging(f: T) -> T:
    """Decorator to log user actions"""

    @functools.wraps(f)
    def wrapper(*args, **kwargs):

        with create_session() as session:
            if g.user.is_anonymous:
                user = 'anonymous'
            else:
                user = g.user.username

            fields_skip_logging = {'csrf_token', '_csrf_token'}
            log = Log(
                event=f.__name__,
                task_instance=None,
                owner=user,
                extra=str([(k, v) for k, v in request.values.items() if k not in fields_skip_logging]),
                task_id=request.values.get('task_id'),
                dag_id=request.values.get('dag_id'),
            )

            if 'execution_date' in request.values:
                log.execution_date = pendulum.parse(request.values.get('execution_date'), strict=False)

            session.add(log)

        return f(*args, **kwargs)

    return cast(T, wrapper)


def gzipped(f: T) -> T:
    """Decorator to make a view compressed"""

    @functools.wraps(f)
    def view_func(*args, **kwargs):
        @after_this_request
        def zipper(response):  # pylint: disable=unused-variable
            accept_encoding = request.headers.get('Accept-Encoding', '')

            if 'gzip' not in accept_encoding.lower():
                return response

            response.direct_passthrough = False

            if (
                response.status_code < 200
                or response.status_code >= 300
                or 'Content-Encoding' in response.headers
            ):
                return response
            gzip_buffer = IO()
            gzip_file = gzip.GzipFile(mode='wb', fileobj=gzip_buffer)
            gzip_file.write(response.data)
            gzip_file.close()

            response.data = gzip_buffer.getvalue()
            response.headers['Content-Encoding'] = 'gzip'
            response.headers['Vary'] = 'Accept-Encoding'
            response.headers['Content-Length'] = len(response.data)

            return response

        return f(*args, **kwargs)

    return cast(T, view_func)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
This module provides helper code to make type annotation within Airflow
codebase easier.
"""

try:
    # Protocol and TypedDict are only added to typing module starting from
    # python 3.8 we can safely remove this shim import after Airflow drops
    # support for <3.8
    from typing import (  # type: ignore # noqa # pylint: disable=unused-import
        Protocol,
        TypedDict,
        runtime_checkable,
    )
except ImportError:
    from typing_extensions import Protocol, TypedDict, runtime_checkable  # type: ignore # noqa



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Classes for interacting with Kubernetes API."""
from abc import ABC, abstractmethod
from functools import reduce
from typing import List, Optional

from kubernetes.client import models as k8s


class K8SModel(ABC):
    """
    These Airflow Kubernetes models are here for backwards compatibility
    reasons only. Ideally clients should use the kubernetes api
    and the process of

        client input -> Airflow k8s models -> k8s models

    can be avoided. All of these models implement the
    `attach_to_pod` method so that they integrate with the kubernetes client.
    """

    @abstractmethod
    def attach_to_pod(self, pod: k8s.V1Pod) -> k8s.V1Pod:
        """
        :param pod: A pod to attach this Kubernetes object to
        :type pod: kubernetes.client.models.V1Pod
        :return: The pod with the object attached
        """


def append_to_pod(pod: k8s.V1Pod, k8s_objects: Optional[List[K8SModel]]):
    """
    :param pod: A pod to attach a list of Kubernetes objects to
    :type pod: kubernetes.client.models.V1Pod
    :param k8s_objects: a potential None list of K8SModels
    :type k8s_objects: Optional[List[K8SModel]]
    :return: pod with the objects attached if they exist
    """
    if not k8s_objects:
        return pod
    return reduce(lambda p, o: o.attach_to_pod(p), k8s_objects, pod)
